from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Events'), name='events')

event_events_exported = namespace.add_event_type(
    label=_('Events exported'), name='event_exported'
)



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .managers import UserOptionsManager


class UserOptions(models.Model):
    """
    This model stores administrative configurations for a user accounts.
    At the moment it stores a boolean flag to restrict a user's
    ability to change their password.
    """
    user = models.OneToOneField(
        on_delete=models.CASCADE, related_name='user_options',
        to=settings.AUTH_USER_MODEL, unique=True, verbose_name=_('User')
    )
    block_password_change = models.BooleanField(
        default=False,
        verbose_name=_('Forbid this user from changing their password.')
    )

    objects = UserOptionsManager()

    class Meta:
        verbose_name = _('User settings')
        verbose_name_plural = _('Users settings')

    def natural_key(self):
        return self.user.natural_key()
    natural_key.dependencies = [settings.AUTH_USER_MODEL]



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.documents.permissions import (
    permission_document_type_edit, permission_document_type_view,
    permission_document_view
)
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import event_index_template_created, event_index_template_edited
from ..models import IndexInstanceNode, IndexTemplate, IndexTemplateNode
from ..permissions import (
    permission_index_template_create, permission_index_template_delete,
    permission_index_template_edit,
    permission_index_instance_view,
    permission_index_template_rebuild, permission_index_template_view
)

from .literals import TEST_INDEX_TEMPLATE_LABEL
from .mixins import (
    DocumentIndexAPIViewTestMixin, IndexInstanceAPIViewTestMixin,
    IndexInstanceNodeAPIViewTestMixin, IndexTemplateTestMixin,
    IndexTemplateActionAPIViewTestMixin, IndexTemplateAPIViewTestMixin,
    IndexTemplateDocumentTypeAPIViewTestMixin,
    IndexTemplateNodeAPITestMixin
)


class DocumentIndexAPIViewTestCase(
    DocumentIndexAPIViewTestMixin, DocumentTestMixin, IndexTemplateTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_index_template(add_test_document_type=True)
        self._create_test_index_template_node(rebuild=True)

    def test_document_index_instance_instance_list_api_view_no_permission(self):
        response = self._request_test_document_index_instance_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_document_index_instance_instance_list_api_view_with_index_access(self):
        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_instance_view
        )

        response = self._request_test_document_index_instance_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_document_index_instance_instance_list_api_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_index_instance_view
        )

        response = self._request_test_document_index_instance_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

    def test_document_index_instance_instance_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document_type,
            permission=permission_index_instance_view
        )
        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_instance_view
        )

        response = self._request_test_document_index_instance_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'], self.test_index_instance_node.pk
        )


class IndexInstanceAPIViewTestCase(
    IndexTemplateTestMixin, IndexInstanceAPIViewTestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_index_template()

    def test_index_instance_detail_api_view_no_permission(self):
        response = self._request_test_index_instance_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('id' not in response.data)

    def test_index_instance_detail_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_instance_view
        )

        response = self._request_test_index_instance_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self.test_index_template.pk
        )

    def test_index_instance_list_api_view_no_permission(self):
        response = self._request_test_index_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

    def test_index_instance_list_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_instance_view
        )

        response = self._request_test_index_instance_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'], self.test_index_template.pk
        )


class IndexInstanceNodeAPIViewTestCase(
    IndexTemplateTestMixin, IndexInstanceNodeAPIViewTestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_index_template(add_test_document_type=True)
        self._create_test_index_template_node(rebuild=True)

    def test_index_instance_node_list_api_view_no_permission(self):
        response = self._request_test_index_instance_node_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response.data)

    def test_index_instance_node_list_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_instance_view
        )

        response = self._request_test_index_instance_node_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'],
            self.test_index_instance_node.pk
        )

    def test_index_instance_node_detail_api_view_no_permission(self):
        response = self._request_test_index_instance_node_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response.data)

    def test_index_instance_node_detail_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_instance_view
        )

        response = self._request_test_index_instance_node_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self.test_index_instance_node.pk
        )

    def test_index_instance_node_document_list_api_view_no_permission(self):
        response = self._request_test_index_instance_node_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response.data)

    def test_index_instance_node_document_list_api_view_with_index_access(self):
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_instance_view
        )

        response = self._request_test_index_instance_node_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

    def test_index_instance_node_document_list_api_view_with_document_access(self):
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_view
        )
        response = self._request_test_index_instance_node_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response.data)

    def test_index_instance_node_document_list_api_view_with_full_access(self):
        self.grant_access(
            obj=self.test_document_type, permission=permission_document_view
        )
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_instance_view
        )

        response = self._request_test_index_instance_node_document_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'], self.test_document.pk
        )


class IndexTemplateAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateAPIViewTestMixin, BaseAPITestCase
):
    def test_index_template_create_api_view_no_permission(self):
        self._clear_events()

        response = self._request_test_index_template_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(IndexTemplate.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_create_api_view_with_permission(self):
        self.grant_permission(permission=permission_index_template_create)

        self._clear_events()

        response = self._request_test_index_template_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.data['id'], self.test_index_template.pk)
        self.assertEqual(response.data['label'], self.test_index_template.label)

        self.assertEqual(IndexTemplate.objects.count(), 1)
        self.assertEqual(
            self.test_index_template.label, TEST_INDEX_TEMPLATE_LABEL
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_index_template)
        self.assertEqual(events[0].verb, event_index_template_created.id)

    def test_index_template_delete_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(self.test_index_template in IndexTemplate.objects.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_delete_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_delete
        )

        self._clear_events()

        response = self._request_test_index_template_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertTrue(self.test_index_template not in IndexTemplate.objects.all())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_detail_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('id' not in response.data)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_detail_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self.test_index_template.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_edit_via_patch_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('id' not in response.data)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_edit_via_patch_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self.test_index_template.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_index_template_list_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_list_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'], self.test_index_template.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexTemplateDocumentTypeAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateDocumentTypeAPIViewTestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def test_index_template_document_type_list_api_view_no_permission(self):
        self._create_test_index_template(add_test_document_type=True)

        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_list_api_view_with_document_type_access(self):
        self._create_test_index_template(add_test_document_type=True)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_view
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_list_api_view_with_index_template_access(self):
        self._create_test_index_template(add_test_document_type=True)

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_list_api_view_with_full_access(self):
        self._create_test_index_template(add_test_document_type=True)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_view
        )
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_view
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'], self.test_document_type.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_no_permission(self):
        self._create_test_index_template()

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self.test_document_type not in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_with_document_type_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self.test_document_type not in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_with_index_template_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(
            self.test_document_type not in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_add_api_view_with_full_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_add_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(
            self.test_document_type in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)

    def test_index_template_document_type_remove_api_view_no_permission(self):
        self._create_test_index_template(add_test_document_type=True)

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self.test_document_type in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_api_view_with_document_type_access(self):
        self._create_test_index_template(add_test_document_type=True)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertTrue(
            self.test_document_type in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_api_view_with_index_template_access(self):
        self._create_test_index_template(add_test_document_type=True)

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertTrue(
            self.test_document_type in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_document_type_remove_api_view_with_full_access(self):
        self._create_test_index_template(add_test_document_type=True)

        self.grant_access(
            obj=self.test_document_type,
            permission=permission_document_type_edit
        )
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )

        self._clear_events()

        response = self._request_test_index_template_document_type_remove_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertTrue(
            self.test_document_type not in self.test_index_template.document_types.all()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self.test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self.test_index_template)
        self.assertEqual(events[0].verb, event_index_template_edited.id)


class IndexTemplateRebuildAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateActionAPIViewTestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._upload_test_document()
        self._create_test_index_template(add_test_document_type=True)
        self._create_test_index_template_node()

    def test_index_template_rebuild_api_view_no_permission(self):
        test_index_instance_node_count = IndexInstanceNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_rebuild_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexInstanceNode.objects.count(), test_index_instance_node_count
        )
        self.assertEqual(IndexInstanceNode.objects.first().parent, None)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_rebuild_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_template_rebuild
        )

        test_index_instance_node_count = IndexInstanceNode.objects.count()

        self._clear_events()

        response = self._request_test_index_template_rebuild_api_view()

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            IndexInstanceNode.objects.count(),
            test_index_instance_node_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexTemplateResetAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateActionAPIViewTestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._create_test_index_template(add_test_document_type=True)
        self._create_test_index_template_node()

        self._create_test_document_stub()
        self.test_index_template.rebuild()

    def test_index_template_reset_api_view_no_permission(self):
        test_index_instance_node_count = self.test_index_template.instance_root.get_children_count()

        self._clear_events()

        response = self._request_test_index_template_reset_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self.test_index_template.instance_root.get_children_count(),
            test_index_instance_node_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_index_template_reset_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template,
            permission=permission_index_template_rebuild
        )

        test_index_instance_node_count = self.test_index_template.instance_root.get_children_count()

        self._clear_events()

        response = self._request_test_index_template_reset_api_view()

        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(
            self.test_index_template.instance_root.get_children_count(),
            test_index_instance_node_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexTemplateNodeAPIViewTestCase(
    IndexTemplateTestMixin, IndexTemplateNodeAPITestMixin, DocumentTestMixin,
    BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_index_template()

    def test_index_template_node_no_parent_create_api_view_no_permission(self):
        index_template_node_count = IndexTemplateNode.objects.count()

        response = self._request_test_index_template_node_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

    def test_index_template_node_no_parent_create_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        response = self._request_test_index_template_node_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

    def test_index_template_node_bad_parent_create_api_view_no_permission(self):
        self._create_test_index_template()

        index_template_node_count = IndexTemplateNode.objects.count()
        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self.test_index_templates[0].template_root.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

    def test_index_template_node_bad_parent_create_api_view_with_access(self):
        self._create_test_index_template()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self.test_index_templates[0].template_root.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

    def test_index_template_node_create_api_view_no_permission(self):
        index_template_node_count = IndexTemplateNode.objects.count()
        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self.test_index_template.template_root.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

    def test_index_template_node_create_api_view_with_access(self):
        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        response = self._request_test_index_template_node_create_api_view(
            extra_data={
                'parent': self.test_index_template.template_root.pk
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(
            response.data['id'], self.test_index_template_node.pk
        )

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count + 1
        )

    def test_index_template_node_delete_api_view_no_permission(self):
        self._create_test_index_template_node()

        index_template_node_count = IndexTemplateNode.objects.count()
        response = self._request_test_index_template_node_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count
        )

    def test_index_template_node_delete_api_view_with_access(self):
        self._create_test_index_template_node()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )
        index_template_node_count = IndexTemplateNode.objects.count()

        response = self._request_test_index_template_node_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            IndexTemplateNode.objects.count(), index_template_node_count - 1
        )

    def test_index_template_node_detail_api_view_no_permission(self):
        self._create_test_index_template_node()

        response = self._request_test_index_template_node_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response.data)

    def test_index_template_node_detail_api_view_with_access(self):
        self._create_test_index_template_node()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_view
        )

        response = self._request_test_index_template_node_detail_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['id'], self.test_index_template_node.pk
        )

    def test_index_template_node_edit_via_patch_api_view_no_permission(self):
        self._create_test_index_template_node()

        index_template_node_expression = self.test_index_template_node.expression

        response = self._request_test_index_template_node_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.test_index_template_node.refresh_from_db()
        self.assertEqual(
            index_template_node_expression,
            self.test_index_template_node.expression
        )

    def test_index_template_node_edit_via_patch_api_view_with_access(self):
        self._create_test_index_template_node()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_edit
        )
        index_template_node_expression = self.test_index_template_node.expression

        response = self._request_test_index_template_node_edit_via_patch_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.test_index_template_node.refresh_from_db()
        self.assertNotEqual(
            index_template_node_expression,
            self.test_index_template_node.expression
        )

    def test_index_template_node_list_api_view_no_permission(self):
        self._create_test_index_template_node()

        response = self._request_test_index_template_node_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertTrue('count' not in response)

    def test_index_template_node_list_api_view_with_access(self):
        self._create_test_index_template_node()

        self.grant_access(
            obj=self.test_index_template, permission=permission_index_template_view
        )

        response = self._request_test_index_template_node_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 1)
        self.assertEqual(
            response.data['results'][0]['id'],
            self.test_index_template_node.pk
        )



import warnings

from django import forms
from django.contrib.auth import authenticate
from django.contrib.auth.forms import AuthenticationForm
from django.forms.widgets import EmailInput
from django.utils.translation import ugettext_lazy as _

from mayan.apps.user_management.querysets import get_user_queryset
from mayan.apps.views.forms import FilteredSelectionForm

from .permissions import permission_users_impersonate


class EmailAuthenticationForm(forms.Form):
    """
    A form to use email address authentication
    """
    email = forms.CharField(
        label=_('Email'), max_length=254, widget=EmailInput()
    )
    password = forms.CharField(
        label=_('Password'), widget=forms.PasswordInput
    )
    remember_me = forms.BooleanField(label=_('Remember me'), required=False)

    error_messages = {
        'invalid_login': _('Please enter a correct email and password. '
                           'Note that the password field is case-sensitive.'),
        'inactive': _('This account is inactive.'),
    }

    def __init__(self, request=None, *args, **kwargs):
        """
        The 'request' parameter is set for custom auth use by subclasses.
        The form data comes in via the standard 'data' kwarg.
        """
        self.request = request
        self.user_cache = None
        super().__init__(*args, **kwargs)

    def clean(self):
        email = self.cleaned_data.get('email')
        password = self.cleaned_data.get('password')

        if email and password:
            self.user_cache = authenticate(email=email, password=password)
            if self.user_cache is None:
                raise forms.ValidationError(
                    self.error_messages['invalid_login'],
                    code='invalid_login',
                )
            elif not self.user_cache.is_active:
                raise forms.ValidationError(
                    self.error_messages['inactive'],
                    code='inactive',
                )
        return self.cleaned_data

    def check_for_test_cookie(self):
        warnings.warn(
            'check_for_test_cookie is deprecated; ensure your login '
            'view is CSRF-protected.', DeprecationWarning
        )

    def get_user_id(self):
        if self.user_cache:
            return self.user_cache.id
        return None

    def get_user(self):
        return self.user_cache


class UserImpersonationOptionsForm(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['permanent'] = forms.BooleanField(
            label=_('Permanent'), help_text=_(
                'If selected, disables ending impersonation.'
            ), required=False
        )


class UserImpersonationSelectionForm(
    FilteredSelectionForm, UserImpersonationOptionsForm
):
    class Meta:
        allow_multiple = False
        field_name = 'user_to_impersonate'
        label = _('User')
        queryset = get_user_queryset().none()
        permission = permission_users_impersonate
        required = True
        widget_attributes = {'class': 'select2'}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        queryset = get_user_queryset().exclude(pk=kwargs['user'].pk)
        self.fields['user_to_impersonate'].queryset = queryset
        self.order_fields(field_order=('user_to_impersonate', 'permanent'))


class UsernameAuthenticationForm(AuthenticationForm):
    remember_me = forms.BooleanField(label=_('Remember me'), required=False)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(
    label=_('Announcements'), name='announcements'
)

permission_announcement_create = namespace.add_permission(
    label=_('Create announcements'), name='announcement_create'
)
permission_announcement_delete = namespace.add_permission(
    label=_('Delete announcements'), name='announcement_delete'
)
permission_announcement_edit = namespace.add_permission(
    label=_('Edit announcements'), name='announcement_edit'
)
permission_announcement_view = namespace.add_permission(
    label=_('View announcements'), name='announcement_view'
)



from django.apps import apps


def method_check_in(self, user=None):
    DocumentCheckout = apps.get_model(
        app_label='checkouts', model_name='DocumentCheckout'
    )

    return DocumentCheckout.business_logic.check_in_document(
        document=self, user=user
    )


def method_get_check_out_info(self):
    DocumentCheckout = apps.get_model(
        app_label='checkouts', model_name='DocumentCheckout'
    )
    return DocumentCheckout.objects.get_check_out_info(document=self)


def method_get_check_out_state(self):
    DocumentCheckout = apps.get_model(
        app_label='checkouts', model_name='DocumentCheckout'
    )
    return DocumentCheckout.objects.get_check_out_state(document=self)


def method_is_checked_out(self):
    DocumentCheckout = apps.get_model(
        app_label='checkouts', model_name='DocumentCheckout'
    )
    return DocumentCheckout.objects.is_checked_out(document=self)



from django import forms
from django.contrib import admindocs
from django.utils.translation import gettext as _
from mayan.apps.common.classes import ModelAttribute
from mayan.apps.views.widgets import NamedMultiWidget

from .classes import Template


class TemplateWidget(NamedMultiWidget):
    builtin_excludes = {
        'tags': ('csrf_token',)
    }

    subwidgets = {
        'builtin_tags': forms.widgets.Select(
            attrs={
                'data-autocopy': 'true',
                'data-field-template': '${ $this.val() }',
            }
        ),
        'template': forms.widgets.Textarea(
            attrs={'rows': 5, 'data-template-fields': 'template'}
        )
    }

    class Media:
        js = ('templating/js/template_widget.js',)

    def get_builtin_choices(self, klass, name_template='{}'):
        result = []
        template = Template('')
        builtin_libraries = [
            ('', library) for library in template._template.engine.template_builtins
        ]
        for module_name, library in builtin_libraries:
            for name, function in getattr(library, klass).items():
                if name not in self.builtin_excludes.get(klass, ()):
                    title, body, metadata = admindocs.utils.parse_docstring(
                        function.__doc__
                    )
                    title = _(title)
                    result.append(
                        (
                            name_template.format(name), '{} - {}'.format(name, title)
                        )
                    )

        result = sorted(result, key=lambda x: x[0])

        return result

    def get_context(self, name, value, attrs):
        result = super().get_context(name=name, value=value, attrs=attrs)
        # Set builtin_tags autocopy sub widget as not required
        result['widget']['subwidgets'][0]['attrs']['required'] = False
        return result

    def decompress(self, value):
        choices_builtin = []
        choices_builtin.append(
            (
                _('Filters'), self.get_builtin_choices(
                    klass='filters', name_template='{{{{ | {} }}}}'
                )
            )
        )
        choices_builtin.append(
            (
                _('Tags'), self.get_builtin_choices(
                    klass='tags', name_template='{{% {} %}}'
                )
            )
        )
        choices_builtin.insert(
            0, ('', _('<Filters and tags>'))
        )

        self.widgets['builtin_tags'].choices = choices_builtin
        return {
            'builtin_tags': None, 'template': value
        }

    def value_from_datadict(self, querydict, files, name):
        template = querydict.get('{}_template'.format(name))

        return template


class ModelTemplateWidget(TemplateWidget):
    def __init__(self, attrs=None, **kwargs):
        super().__init__(attrs=attrs, **kwargs)
        self.widgets['model_attribute'] = forms.widgets.Select(
            attrs={
                'data-autocopy': 'true',
                'data-field-template': '{{ ${ $idTemplate.data("model-variable") }.${ $this.val() } }}'
            }
        )
        self.subwidgets_order.insert(0, 'model_attribute')

    def get_context(self, name, value, attrs):
        result = super().get_context(name=name, value=value, attrs=attrs)
        # Set model_attribute autocopy sub widget as not required
        result['widget']['subwidgets'][1]['attrs']['required'] = False
        return result

    def decompress(self, value):
        result = super().decompress(value=value)

        attribute_choices = ModelAttribute.get_all_choices_for(
            model=self.attrs['model']
        )
        attribute_choices.insert(
            0, ('', _('<Model attributes>'))
        )

        self.widgets['model_attribute'].choices = attribute_choices

        result['model_attribute'] = None

        return result



from django.core.exceptions import ValidationError

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.testing.tests.base import BaseTestCase

from ..models import Cabinet

from .literals import TEST_CABINET_LABEL
from .mixins import CabinetTestMixin


class CabinetTestCase(CabinetTestMixin, BaseTestCase):
    def test_cabinet_creation(self):
        self._create_test_cabinet()

        self.assertEqual(Cabinet.objects.all().count(), 1)
        self.assertQuerysetEqual(
            Cabinet.objects.all(), (repr(self.test_cabinet),)
        )

    def test_cabinet_duplicate_creation(self):
        self._create_test_cabinet()

        with self.assertRaises(expected_exception=ValidationError):
            cabinet_2 = Cabinet(label=TEST_CABINET_LABEL)
            cabinet_2.validate_unique()
            cabinet_2.save()

        self.assertEqual(Cabinet.objects.all().count(), 1)
        self.assertQuerysetEqual(
            Cabinet.objects.all(), (repr(self.test_cabinet),)
        )

    def test_inner_cabinet_creation(self):
        self._create_test_cabinet()

        inner_cabinet = Cabinet.objects.create(
            parent=self.test_cabinet, label=TEST_CABINET_LABEL
        )

        self.assertEqual(Cabinet.objects.all().count(), 2)
        self.assertQuerysetEqual(
            Cabinet.objects.all(),
            map(repr, (self.test_cabinet, inner_cabinet))
        )

    def test_method_get_absolute_url(self):
        self._create_test_cabinet()

        self.assertTrue(self.test_cabinet.get_absolute_url())


class CabinetDocumentTestCase(
    CabinetTestMixin, DocumentTestMixin, BaseTestCase
):
    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_cabinet()

    def test_addition_of_documents(self):
        self.test_cabinet.documents.add(self.test_document)

        self.assertEqual(self.test_cabinet.documents.count(), 1)
        self.assertQuerysetEqual(
            self.test_cabinet.documents.all(), (repr(self.test_document),)
        )

    def test_addition_and_deletion_of_documents(self):
        self.test_cabinet.documents.add(self.test_document)

        self.assertEqual(self.test_cabinet.documents.count(), 1)
        self.assertQuerysetEqual(
            self.test_cabinet.documents.all(), (repr(self.test_document),)
        )

        self.test_cabinet.documents.remove(self.test_document)

        self.assertEqual(self.test_cabinet.documents.count(), 0)
        self.assertQuerysetEqual(self.test_cabinet.documents.all(), ())



from django.utils.translation import ugettext_lazy as _

from mayan.apps.dependencies.classes import BinaryDependency

from .drivers.exiftool import EXIFToolDriver

exiftool = EXIFToolDriver(auto_initialize=False)
exiftool.read_settings()

BinaryDependency(
    help_text=_(
        'Library and program to read and write meta information in multimedia '
        'files.'
    ), module=__name__, name='exiftool', path=exiftool.exiftool_path
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import get_cascade_condition

from .icons import (
    icon_document_file_content, icon_document_file_content_delete,
    icon_document_file_content_download, icon_document_file_page_content,
    icon_document_file_parsing_errors_list, icon_document_file_submit,
    icon_document_type_parsing_settings, icon_document_type_submit,
    icon_error_list
)
from .permissions import (
    permission_document_file_content_view, permission_document_type_parsing_setup,
    permission_document_file_parse
)


link_document_file_content = Link(
    args='resolved_object.id', icon=icon_document_file_content,
    permissions=(permission_document_file_content_view,), text=_('Content'),
    view='document_parsing:document_file_content_view'
)
link_document_file_content_delete = Link(
    args='resolved_object.id', icon=icon_document_file_content_delete,
    permissions=(permission_document_file_parse,),
    text=_('Delete parsed content'),
    view='document_parsing:document_file_content_delete',
)
link_document_file_multiple_content_delete = Link(
    icon=icon_document_file_content_delete, text=_('Delete parsed content'),
    view='document_parsing:document_file_multiple_content_delete',
)
link_document_file_content_download = Link(
    args='resolved_object.id',
    icon=icon_document_file_content_download,
    permissions=(permission_document_file_content_view,), text=_('Download content'),
    view='document_parsing:document_file_content_download'
)
link_document_file_page_content = Link(
    args='resolved_object.id',
    icon=icon_document_file_page_content,
    permissions=(permission_document_file_content_view,), text=_('Content'),
    view='document_parsing:document_file_page_content_view'
)
link_document_file_parsing_errors_list = Link(
    args='resolved_object.id',
    icon=icon_document_file_parsing_errors_list,
    permissions=(permission_document_file_parse,), text=_('Parsing errors'),
    view='document_parsing:document_file_parsing_error_list'
)
link_document_file_multiple_submit = Link(
    icon=icon_document_file_submit,
    text=_('Submit for parsing'),
    view='document_parsing:document_file_multiple_submit'
)
link_document_file_submit = Link(
    args='resolved_object.id',
    icon=icon_document_file_submit,
    permissions=(permission_document_file_parse,),
    text=_('Submit for parsing'),
    view='document_parsing:document_file_submit'
)
link_document_type_parsing_settings = Link(
    args='resolved_object.id',
    icon=icon_document_type_parsing_settings,
    permissions=(permission_document_type_parsing_setup,),
    text=_('Setup parsing'),
    view='document_parsing:document_type_parsing_settings'
)
link_document_type_submit = Link(
    condition=get_cascade_condition(
        app_label='documents', model_name='DocumentType',
        object_permission=permission_document_type_parsing_setup
    ),
    icon=icon_document_type_submit,
    text=_('Parse documents per type'),
    view='document_parsing:document_type_submit'
)
link_error_list = Link(
    icon=icon_error_list,
    permissions=(permission_document_file_parse,), text=_('Parsing errors'),
    view='document_parsing:error_list'
)



from mayan.apps.appearance.classes import Icon
from mayan.apps.documents.icons import icon_document_type

icon_document_metadata_add = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='plus'
)
icon_document_metadata_edit = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='pen'
)
icon_document_metadata_remove = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='minus'
)
icon_document_metadata_view = Icon(
    driver_name='fontawesome', symbol='database'
)
icon_document_multiple_metadata_add = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='plus'
)
icon_document_multiple_metadata_edit = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='pen'
)
icon_document_multiple_metadata_remove = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='minus'
)
icon_document_type_metadata_type_list = Icon(
    driver_name='fontawesome', symbol='database'
)
icon_metadata = Icon(driver_name='fontawesome', symbol='database')
icon_metadata_type_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='database',
    secondary_symbol='plus'
)
icon_metadata_type_delete = Icon(driver_name='fontawesome', symbol='times')
icon_metadata_type_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_metadata_type_list = Icon(driver_name='fontawesome', symbol='database')
icon_metadata_type_document_type_list = icon_document_type



from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers
from rest_framework.reverse import reverse

from mayan.apps.django_gpg.models import Key
from mayan.apps.django_gpg.permissions import permission_key_sign
from mayan.apps.rest_api.relations import (
    FilteredPrimaryKeyRelatedField, MultiKwargHyperlinkedIdentityField
)

from .models import DetachedSignature, EmbeddedSignature


class BaseSignatureSerializer(serializers.HyperlinkedModelSerializer):
    document_file_url = MultiKwargHyperlinkedIdentityField(
        view_kwargs=(
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id',
            },
            {
                'lookup_field': 'document_file.document_id',
                'lookup_url_kwarg': 'document_id',
            }
        ),
        view_name='rest_api:documentfile-detail'
    )
    key_url = serializers.SerializerMethodField()

    class Meta:
        fields = (
            'date_time', 'document_file_url', 'key_algorithm',
            'key_creation_date', 'key_expiration_date', 'key_id',
            'key_length', 'key_type', 'key_user_id', 'key_url',
            'public_key_fingerprint', 'signature_id', 'url'
        )
        read_only_fields = (
            'date_time', 'document_file_url', 'key_algorithm',
            'key_creation_date', 'key_expiration_date', 'key_id',
            'key_length', 'key_type', 'key_user_id', 'key_url',
            'public_key_fingerprint', 'signature_id', 'url'
        )

    def get_key_url(self, instance):
        key = instance.key

        if key:
            return reverse(
                viewname='rest_api:key-detail', kwargs={
                    'key_id': key.pk
                }, request=self.context['request'],
                format=self.context['format']
            )


class BaseSignSerializer(serializers.HyperlinkedModelSerializer):
    key = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the tag to add to the document.'
        ), source_queryset=Key.objects.private_keys(),
        source_permission=permission_key_sign
    )
    passphrase = serializers.CharField(
        help_text=_(
            'The passphrase to unlock the key and allow it to be used to '
            'sign the document file.'
        ),
        required=False, write_only=True
    )

    class Meta:
        fields = ('key', 'passphrase',)


class DetachedSignatureSerializer(BaseSignatureSerializer):
    url = MultiKwargHyperlinkedIdentityField(
        view_kwargs=(
            {
                'lookup_field': 'document_file.document_id',
                'lookup_url_kwarg': 'document_id',
            },
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id',
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'detached_signature_id',
            },
        ),
        view_name='rest_api:detachedsignature-detail'
    )

    class Meta(BaseSignatureSerializer.Meta):
        model = DetachedSignature


class DetachedSignatureUploadSerializer(DetachedSignatureSerializer):
    class Meta(DetachedSignatureSerializer.Meta):
        fields = DetachedSignatureSerializer.Meta.fields + ('signature_file',)
        model = DetachedSignature


class EmbeddedSignatureSerializer(
    BaseSignatureSerializer, serializers.HyperlinkedModelSerializer
):
    url = MultiKwargHyperlinkedIdentityField(
        view_kwargs=(
            {
                'lookup_field': 'document_file.document.pk',
                'lookup_url_kwarg': 'document_id',
            },
            {
                'lookup_field': 'document_file_id',
                'lookup_url_kwarg': 'document_file_id',
            },
            {
                'lookup_field': 'pk',
                'lookup_url_kwarg': 'embedded_signature_id',
            },
        ),
        view_name='rest_api:embeddedsignature-detail'
    )

    class Meta(BaseSignatureSerializer.Meta):
        model = EmbeddedSignature


class SignDetachedSerializer(BaseSignSerializer, BaseSignatureSerializer):
    class Meta(BaseSignSerializer.Meta, BaseSignatureSerializer.Meta):
        model = DetachedSignature


class SignEmbeddedSerializer(SignDetachedSerializer):
    class Meta(SignDetachedSerializer.Meta):
        model = EmbeddedSignature



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0003_auto_20150608_1915'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='document',
            options={
                'ordering': ('-date_added',), 'verbose_name': 'Document',
                'verbose_name_plural': 'Documents'
            },
        ),
        migrations.AlterModelOptions(
            name='documentpage',
            options={
                'ordering': ('page_number',), 'verbose_name': 'Document page',
                'verbose_name_plural': 'Document pages'
            },
        ),
        migrations.AlterModelOptions(
            name='documenttype',
            options={
                'ordering': ('name',), 'verbose_name': 'Document type',
                'verbose_name_plural': 'Documents types'
            },
        ),
        migrations.AlterModelOptions(
            name='documenttypefilename',
            options={
                'ordering': ('filename',),
                'verbose_name': 'Document type quick rename filename',
                'verbose_name_plural': 'Document types quick rename filenames'
            },
        ),
    ]




from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_SOURCES_SCANIMAGE_PATH,
    DEFAULT_SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND,
    DEFAULT_SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND_ARGUMENTS
)
from .setting_migrations import SourcesSettingMigration

namespace = SettingNamespace(
    label=_('Sources'), migration_class=SourcesSettingMigration,
    name='sources', version='0002'
)

setting_scanimage_path = namespace.add_setting(
    default=DEFAULT_SOURCES_SCANIMAGE_PATH,
    global_name='SOURCES_SCANIMAGE_PATH', help_text=_(
        'File path to the scanimage program used to control image scanners.'
    ), is_path=True
)
setting_staging_file_image_cache_storage = namespace.add_setting(
    global_name='SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND',
    default=DEFAULT_SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND, help_text=_(
        'Path to the Storage subclass to use when storing the cached '
        'staging_file image files.'
    )
)
setting_staging_file_image_cache_storage_arguments = namespace.add_setting(
    global_name='SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND_ARGUMENTS',
    default=DEFAULT_SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND_ARGUMENTS,
    help_text=_(
        'Arguments to pass to the SOURCES_STAGING_FILE_CACHE_STORAGE_BACKEND.'
    )
)



from django.core import management

from ...backends.base import LockingBackend


class Command(management.BaseCommand):
    help = 'Erase all locks (acquired and stale).'

    def handle(self, *args, **options):
        LockingBackend.get_backend().purge_locks()



from ..literals import USER_IMPERSONATE_VARIABLE_ID


class UserPasswordViewTestMixin:
    def _request_test_user_password_set_view(self, password):
        return self.post(
            viewname='authentication:user_set_password', kwargs={
                'user_id': self.test_user.pk
            }, data={
                'new_password1': password, 'new_password2': password
            }
        )

    def _request_test_user_password_set_multiple_view(self, password):
        return self.post(
            viewname='authentication:user_multiple_set_password', data={
                'id_list': self.test_user.pk,
                'new_password1': password,
                'new_password2': password
            }
        )


class UserImpersonationViewTestMixin:
    def _impersonate_test_user(self):
        session = self.client.session
        session[USER_IMPERSONATE_VARIABLE_ID] = self.test_user.pk
        session.save()

    def _request_test_user_impersonate_end_view(self):
        return self.get(
            follow=True, viewname='authentication:user_impersonate_end'
        )

    def _request_test_user_impersonate_form_start_view(self):
        return self.post(
            follow=True, viewname='authentication:user_impersonate_form_start',
            data={
                'user_to_impersonate': self.test_user.pk
            }
        )

    def _request_test_user_impersonate_start_view(self):
        return self.post(
            follow=True, viewname='authentication:user_impersonate_start', kwargs={
                'user_id': self.test_user.pk
            }
        )


class UserLoginTestMixin:
    def _request_authenticated_view(self):
        return self.get(path=self.authenticated_url)

    def _request_password_reset_get_view(self):
        return self.get(
            viewname='authentication:password_reset_view', data={
                'email': self._test_case_superuser.email,
            }
        )

    def _request_password_reset_post_view(self):
        return self.post(
            viewname='authentication:password_reset_view', data={
                'email': self._test_case_superuser.email,
            }
        )



from actstream.models import Action

from ..events import event_document_trashed, event_document_viewed
from ..permissions import permission_document_trash, permission_document_view

from .base import GenericDocumentViewTestCase
from .mixins.document_mixins import DocumentViewTestMixin
from .mixins.trashed_document_mixins import TrashedDocumentViewTestMixin


class DocumentEventsTestCase(
    DocumentViewTestMixin, TrashedDocumentViewTestMixin,
    GenericDocumentViewTestCase
):
    def setUp(self):
        super().setUp()
        Action.objects.all().delete()

    def test_document_view_event_no_permission(self):
        response = self._request_test_document_preview_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(list(Action.objects.any(obj=self.test_document)), [])

    def test_document_view_event_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_view
        )

        response = self._request_test_document_preview_view()
        self.assertEqual(response.status_code, 200)

        event = Action.objects.any(obj=self.test_document).first()
        self.assertEqual(event.actor, self._test_case_user)
        self.assertEqual(event.target, self.test_document)
        self.assertEqual(event.verb, event_document_viewed.id)

    def test_document_trashed_view_event_no_permission(self):
        response = self._request_test_document_trash_post_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(list(Action.objects.any(obj=self.test_document)), [])

    def test_document_trashed_view_event_with_access(self):
        self.grant_access(
            obj=self.test_document, permission=permission_document_trash
        )

        response = self._request_test_document_trash_post_view()
        self.assertEqual(response.status_code, 302)

        event = Action.objects.any(obj=self.test_document).first()
        self.assertEqual(event.actor, self._test_case_user)
        self.assertEqual(event.target, self.test_document)
        self.assertEqual(event.verb, event_document_trashed.id)



from mayan.apps.appearance.classes import Icon

icon_about = Icon(driver_name='fontawesome', symbol='info')
icon_book = Icon(driver_name='fontawesome', symbol='book')
icon_documentation = Icon(driver_name='fontawesome', symbol='book')
icon_forum = Icon(
    driver_name='fontawesome', symbol='life-ring'
)
icon_license = Icon(
    driver_name='fontawesome', symbol='certificate'
)
icon_menu_about = Icon(
    driver_name='fontawesome', symbol='info'
)
icon_menu_user = Icon(
    driver_name='fontawesome', symbol='user-circle'
)
icon_object_copy = Icon(
    driver_name='fontawesome', symbol='stamp'
)
icon_setup = Icon(
    driver_name='fontawesome', symbol='cog'
)
icon_social_paypal = Icon(
    driver_name='fontawesomecss', css_classes='fab fa-paypal'
)
icon_source_code = Icon(driver_name='fontawesome', symbol='code-branch')
icon_store = Icon(driver_name='fontawesome', symbol='tshirt')
icon_support = Icon(
    driver_name='fontawesome', symbol='phone'
)
icon_tools = Icon(
    driver_name='fontawesome', symbol='wrench'
)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.smart_settings.classes import SettingNamespace

from .literals import (
    DEFAULT_CONVERTER_ASSET_CACHE_MAXIMUM_SIZE,
    DEFAULT_CONVERTER_ASSET_CACHE_TIME,
    DEFAULT_CONVERTER_ASSET_CACHE_STORAGE_BACKEND,
    DEFAULT_CONVERTER_ASSET_CACHE_STORAGE_BACKEND_ARGUMENTS,
    DEFAULT_CONVERTER_ASSET_STORAGE_BACKEND,
    DEFAULT_CONVERTER_ASSET_STORAGE_BACKEND_ARGUMENTS,
    DEFAULT_CONVERTER_GRAPHICS_BACKEND,
    DEFAULT_CONVERTER_GRAPHICS_BACKEND_ARGUMENTS
)
from .setting_callbacks import callback_update_asset_cache_size
from .setting_migrations import ConvertSettingMigration

namespace = SettingNamespace(
    label=_('Converter'), migration_class=ConvertSettingMigration,
    name='converter', version='0002'
)


setting_asset_cache_maximum_size = namespace.add_setting(
    default=DEFAULT_CONVERTER_ASSET_CACHE_MAXIMUM_SIZE,
    global_name='CONVERTER_ASSET_CACHE_MAXIMUM_SIZE',
    help_text=_(
        'The threshold at which the CONVERTER_ASSET_CACHE_STORAGE_BACKEND '
        'will start deleting the oldest asset cache files. '
        'Specify the size in bytes.'
    ), post_edit_function=callback_update_asset_cache_size
)
setting_asset_cache_time = namespace.add_setting(
    default=DEFAULT_CONVERTER_ASSET_CACHE_TIME,
    global_name='CONVERTER_ASSET_CACHE_TIME',
    help_text=_(
        'Time in seconds that the browser should cache the supplied asset. '
        'The default of 31559626 seconds correspond to 1 year.'
    )
)
setting_asset_cache_storage_backend = namespace.add_setting(
    default=DEFAULT_CONVERTER_ASSET_CACHE_STORAGE_BACKEND,
    global_name='CONVERTER_ASSET_CACHE_STORAGE_BACKEND', help_text=_(
        'Path to the Storage subclass to use when storing the cached '
        'asset files.'
    )
)
setting_asset_cache_storage_backend_arguments = namespace.add_setting(
    default=DEFAULT_CONVERTER_ASSET_CACHE_STORAGE_BACKEND_ARGUMENTS,
    global_name='CONVERTER_ASSET_CACHE_STORAGE_BACKEND_ARGUMENTS',
    help_text=_(
        'Arguments to pass to the CONVERTER_ASSET_CACHE_STORAGE_BACKEND.'
    )
)
setting_asset_storage_backend = namespace.add_setting(
    default=DEFAULT_CONVERTER_ASSET_STORAGE_BACKEND,
    global_name='CONVERTER_ASSET_STORAGE_BACKEND', help_text=_(
        'Path to the Storage subclass to use when storing assets.'
    )
)
setting_asset_storage_backend_arguments = namespace.add_setting(
    default=DEFAULT_CONVERTER_ASSET_STORAGE_BACKEND_ARGUMENTS,
    global_name='CONVERTER_ASSET_STORAGE_BACKEND_ARGUMENTS', help_text=_(
        'Arguments to pass to the CONVERTER_ASSET_STORAGE_BACKEND.'
    )
)
setting_graphics_backend = namespace.add_setting(
    default=DEFAULT_CONVERTER_GRAPHICS_BACKEND,
    global_name='CONVERTER_GRAPHICS_BACKEND', help_text=_(
        'Graphics conversion backend to use.'
    )
)
setting_graphics_backend_arguments = namespace.add_setting(
    default=DEFAULT_CONVERTER_GRAPHICS_BACKEND_ARGUMENTS,
    global_name='CONVERTER_GRAPHICS_BACKEND_ARGUMENTS', help_text=_(
        'Configuration options for the graphics conversion backend.'
    )
)



from django.db.models import Q

from ..models import Message

from .literals import TEST_MESSAGE_BODY, TEST_MESSAGE_SUBJECT


class MessageTestMixin:
    def _create_test_message(self):
        self.test_message = Message.objects.create(
            body=TEST_MESSAGE_BODY, subject=TEST_MESSAGE_SUBJECT,
            user=self._test_case_user
        )


class MessageViewTestMixin:
    def _request_test_message_create_view(self):
        pk_list = list(Message.objects.values('pk'))

        response = self.post(
            viewname='messaging:message_create', data={
                'body': TEST_MESSAGE_BODY,
                'subject': TEST_MESSAGE_SUBJECT,
                'user': self._test_case_user.pk
            }
        )

        try:
            self.test_message = Message.objects.get(~Q(pk__in=pk_list))
        except Message.DoesNotExist:
            self.test_message = None

        return response

    def _request_test_message_delete_view(self):
        return self.post(
            viewname='messaging:message_single_delete', kwargs={
                'message_id': self.test_message.pk
            }
        )

    def _request_test_message_detail_view(self):
        return self.get(
            viewname='messaging:message_detail', kwargs={
                'message_id': self.test_message.pk
            }
        )

    def _request_test_message_list_view(self):
        return self.get(viewname='messaging:message_list')



from django.core import management
from django.utils.encoding import force_text

from mayan.apps.documents.tests.base import GenericDocumentTestCase
from mayan.apps.documents.storages import storage_document_files
from mayan.apps.mimetype.api import get_mimetype

from .mixins import StorageProcessorTestMixin


class StorageProcessManagementCommandTestCase(
    StorageProcessorTestMixin, GenericDocumentTestCase
):
    def _call_command(self, reverse=None):
        options = {
            'app_label': 'documents',
            'defined_storage_name': storage_document_files.name,
            'log_file': force_text(s=self.path_test_file),
            'model_name': 'DocumentFile',
            'reverse': reverse
        }
        management.call_command(command_name='storage_process', **options)

    def _upload_and_call(self):
        self.defined_storage.dotted_path = 'django.core.files.storage.FileSystemStorage'
        self.defined_storage.kwargs = {
            'location': self.document_storage_kwargs['location']
        }

        self._upload_test_document()

        self.defined_storage.dotted_path = 'mayan.apps.storage.backends.compressedstorage.ZipCompressedPassthroughStorage'
        self.defined_storage.kwargs = {
            'next_storage_backend': 'django.core.files.storage.FileSystemStorage',
            'next_storage_backend_arguments': {
                'location': self.document_storage_kwargs['location']
            }
        }

        self._call_command()

    def test_storage_processor_command_forwards(self):
        self._upload_and_call()

        with open(file=self.test_document.file_latest.file.path, mode='rb') as file_object:
            self.assertEqual(
                get_mimetype(file_object=file_object),
                ('application/zip', 'binary')
            )

        self.assertEqual(
            self.test_document.file_latest.checksum,
            self.test_document.file_latest.checksum_update(save=False)
        )

    def test_processor_forwards_and_reverse(self):
        self._upload_and_call()

        self._call_command(reverse=True)

        self.defined_storage.dotted_path = 'django.core.files.storage.FileSystemStorage'
        self.defined_storage.kwargs = {
            'location': self.document_storage_kwargs['location']
        }

        with open(file=self.test_document.file_latest.file.path, mode='rb') as file_object:
            self.assertNotEqual(
                get_mimetype(file_object=file_object),
                ('application/zip', 'binary')
            )

        self.assertEqual(
            self.test_document.file_latest.checksum,
            self.test_document.file_latest.checksum_update(save=False)
        )



from django.apps import apps
from django.urls import reverse_lazy
from django.utils.translation import ugettext_lazy as _

from mayan.apps.dashboards.classes import DashboardWidgetNumeric
from mayan.apps.mayan_statistics.icons import icon_statistics

from .icons import (
    icon_dashboard_documents_in_trash, icon_dashboard_document_types,
    icon_dashboard_pages_per_month, icon_dashboard_new_documents_this_month,
    icon_dashboard_total_document
)
from .permissions import (
    permission_document_view, permission_document_type_view
)
from .statistics import (
    new_document_pages_this_month, new_documents_this_month,
)


class DashboardWidgetDocumentFilePagesTotal(DashboardWidgetNumeric):
    icon = icon_dashboard_pages_per_month
    label = _('Total pages')
    link = reverse_lazy(
        viewname='statistics:statistic_detail', kwargs={
            'slug': 'total-document-pages-at-each-month'
        }
    )
    link_icon = icon_statistics

    def render(self, request):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        DocumentFilePage = apps.get_model(
            app_label='documents', model_name='DocumentFilePage'
        )
        self.count = AccessControlList.objects.restrict_queryset(
            permission=permission_document_view, user=request.user,
            queryset=DocumentFilePage.valid.all()
        ).count()
        return super().render(request)


class DashboardWidgetDocumentsTotal(DashboardWidgetNumeric):
    icon = icon_dashboard_total_document
    label = _('Total documents')
    link = reverse_lazy(viewname='documents:document_list')

    def render(self, request):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        Document = apps.get_model(
            app_label='documents', model_name='Document'
        )
        self.count = AccessControlList.objects.restrict_queryset(
            permission=permission_document_view, user=request.user,
            queryset=Document.valid.all()
        ).count()
        return super().render(request)


class DashboardWidgetDocumentsInTrash(DashboardWidgetNumeric):
    icon = icon_dashboard_documents_in_trash
    label = _('Documents in trash')
    link = reverse_lazy(viewname='documents:document_list_deleted')

    def render(self, request):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        TrashedDocument = apps.get_model(
            app_label='documents', model_name='TrashedDocument'
        )
        self.count = AccessControlList.objects.restrict_queryset(
            permission=permission_document_view, user=request.user,
            queryset=TrashedDocument.objects.all()
        ).count()
        return super().render(request)


class DashboardWidgetDocumentsTypesTotal(DashboardWidgetNumeric):
    icon = icon_dashboard_document_types
    label = _('Document types')
    link = reverse_lazy(viewname='documents:document_type_list')

    def render(self, request):
        AccessControlList = apps.get_model(
            app_label='acls', model_name='AccessControlList'
        )
        DocumentType = apps.get_model(
            app_label='documents', model_name='DocumentType'
        )
        self.count = AccessControlList.objects.restrict_queryset(
            permission=permission_document_type_view, user=request.user,
            queryset=DocumentType.objects.all()
        ).count()
        return super().render(request)


class DashboardWidgetDocumentsNewThisMonth(DashboardWidgetNumeric):
    icon = icon_dashboard_new_documents_this_month
    label = _('New documents this month')
    link = reverse_lazy(
        viewname='statistics:statistic_detail', kwargs={
            'slug': 'new-documents-per-month'
        }
    )
    link_icon = icon_statistics

    def render(self, request):
        self.count = new_documents_this_month(user=request.user)
        return super().render(request)


class DashboardWidgetDocumentsPagesNewThisMonth(DashboardWidgetNumeric):
    icon = icon_dashboard_pages_per_month
    label = _('New pages this month')
    link = reverse_lazy(
        viewname='statistics:statistic_detail', kwargs={
            'slug': 'new-document-pages-per-month'
        }
    )
    link_icon = icon_statistics

    def render(self, request):
        self.count = new_document_pages_this_month(user=request.user)
        return super().render(request)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(label=_('Metadata'), name='metadata')

permission_document_metadata_add = namespace.add_permission(
    label=_('Add metadata to a document'), name='metadata_document_add'
)
permission_document_metadata_edit = namespace.add_permission(
    label=_('Edit a document\'s metadata'), name='metadata_document_edit'
)
permission_document_metadata_remove = namespace.add_permission(
    label=_('Remove metadata from a document'),
    name='metadata_document_remove'
)
permission_document_metadata_view = namespace.add_permission(
    label=_('View metadata from a document'), name='metadata_document_view'
)

setup_namespace = PermissionNamespace(
    label=_('Metadata setup'), name='metadata_setup'
)

permission_metadata_type_edit = setup_namespace.add_permission(
    label=_('Edit metadata types'), name='metadata_type_edit'
)
permission_metadata_type_create = setup_namespace.add_permission(
    label=_('Create new metadata types'), name='metadata_type_create'
)
permission_metadata_type_delete = setup_namespace.add_permission(
    label=_('Delete metadata types'), name='metadata_type_delete'
)
permission_metadata_type_view = setup_namespace.add_permission(
    label=_('View metadata types'), name='metadata_type_view'
)



from django.core.exceptions import ImproperlyConfigured
from django.db.models import Q

from mayan.apps.permissions.tests.mixins import (
    RoleTestCaseMixin, RoleTestMixin
)
from mayan.apps.user_management.tests.mixins import UserTestCaseMixin

from ..classes import ModelPermission
from ..models import AccessControlList
from ..permissions import permission_acl_edit, permission_acl_view


class ACLAPIViewTestMixin:
    def _request_test_acl_create_api_view(self):
        pk_list = list(AccessControlList.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='rest_api:accesscontrollist-list',
            kwargs=self.test_object_view_kwargs, data={
                'role_id': self.test_role.pk
            }
        )

        try:
            self.test_acl = AccessControlList.objects.get(
                ~Q(pk__in=pk_list)
            )
        except AccessControlList.DoesNotExist:
            self.test_acl = None

        return response

    def _request_test_acl_delete_api_view(self):
        return self.delete(
            viewname='rest_api:accesscontrollist-detail', kwargs={
                'app_label': self.test_object_content_type.app_label,
                'model_name': self.test_object_content_type.model,
                'object_id': self.test_object.pk,
                'acl_id': self.test_acl.pk
            }
        )

    def _request_test_acl_detail_api_view(self):
        return self.get(
            viewname='rest_api:accesscontrollist-detail', kwargs={
                'app_label': self.test_object_content_type.app_label,
                'model_name': self.test_object_content_type.model,
                'object_id': self.test_object.pk,
                'acl_id': self.test_acl.pk
            }
        )

    def _request_test_acl_list_api_view(self):
        return self.get(
            viewname='rest_api:accesscontrollist-list', kwargs={
                'app_label': self.test_object_content_type.app_label,
                'model_name': self.test_object_content_type.model,
                'object_id': self.test_object.pk
            }
        )

    def _request_test_acl_permission_add_api_view(self):
        return self.post(
            viewname='rest_api:accesscontrollist-permission-add', kwargs={
                'app_label': self.test_object_content_type.app_label,
                'model_name': self.test_object_content_type.model,
                'object_id': self.test_object.pk,
                'acl_id': self.test_acl.pk
            }, data={'permission': self.test_permission.pk}
        )

    def _request_test_acl_permission_list_api_view(self):
        return self.get(
            viewname='rest_api:accesscontrollist-permission-list', kwargs={
                'app_label': self.test_object_content_type.app_label,
                'model_name': self.test_object_content_type.model,
                'object_id': self.test_object.pk,
                'acl_id': self.test_acl.pk
            }
        )

    def _request_test_acl_permission_remove_api_view(self):
        return self.post(
            viewname='rest_api:accesscontrollist-permission-remove', kwargs={
                'app_label': self.test_object_content_type.app_label,
                'model_name': self.test_object_content_type.model,
                'object_id': self.test_object.pk,
                'acl_id': self.test_acl.pk
            }, data={'permission': self.test_permission.pk}
        )

    def _request_test_class_permission_list_api_view(self):
        return self.get(
            viewname='rest_api:class-permission-list', kwargs={
                'app_label': self.test_object_content_type.app_label,
                'model_name': self.test_object_content_type.model,
            }
        )


class ACLTestCaseMixin(RoleTestCaseMixin, UserTestCaseMixin):
    def setUp(self):
        super().setUp()
        if hasattr(self, '_test_case_user'):
            self._test_case_role.groups.add(self._test_case_group)

    def grant_access(self, obj, permission):
        if not hasattr(self, '_test_case_role'):
            raise ImproperlyConfigured(
                'Enable the creation of the test case user, group, and role '
                'in order to enable the usage of ACLs in tests.'
            )

        self._test_case_acl = AccessControlList.objects.grant(
            obj=obj, permission=permission, role=self._test_case_role
        )

    def revoke_access(self, obj, permission):
        self._test_case_acl = AccessControlList.objects.revoke(
            obj=obj, permission=permission, role=self._test_case_role
        )


class ACLTestMixin(RoleTestMixin):
    auto_create_test_role = True
    auto_create_acl_test_object = False

    def setUp(self):
        super().setUp()
        if self.auto_create_test_role:
            self._create_test_role()

        if self.auto_create_acl_test_object:
            self._create_acl_test_object()

    def _create_test_acl(self):
        self.test_acl = AccessControlList.objects.create(
            content_object=self.test_object, role=self.test_role
        )

    def _create_acl_test_object(
        self, model_name=None, create_test_permission=True,
        register_model_permissions=True
    ):
        self._create_test_object(
            create_test_permission=create_test_permission
        )

        if register_model_permissions:
            ModelPermission.register(
                model=self.TestModel, permissions=(
                    permission_acl_edit, permission_acl_view,
                )
            )

    def _create_acl_test_object_base(self):
        self.test_object_base = self._create_acl_test_object(
            model_name='TestModelBase'
        )

    def _create_acl_test_object_proxy(self):
        self.TestModelProxy = self._create_test_model(
            base_class=self.TestModel, model_name='TestModelProxy',
            options={
                'proxy': True
            }
        )
        self.test_object_proxy = self.TestModelProxy.objects.create()


class AccessControlListViewTestMixin:
    def _request_test_acl_create_get_view(self):
        return self.get(
            viewname='acls:acl_create',
            kwargs=self.test_object_view_kwargs, data={
                'role': self.test_role.pk
            }
        )

    def _request_test_acl_create_post_view(self):
        pk_list = list(AccessControlList.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='acls:acl_create',
            kwargs=self.test_object_view_kwargs, data={
                'role': self.test_role.pk
            }
        )

        try:
            self.test_acl = AccessControlList.objects.get(
                ~Q(pk__in=pk_list)
            )
        except AccessControlList.DoesNotExist:
            self.test_acl = None

        return response

    def _request_test_acl_delete_view(self):
        return self.post(
            viewname='acls:acl_delete', kwargs={'acl_id': self.test_acl.pk}
        )

    def _request_test_acl_list_view(self):
        return self.get(
            viewname='acls:acl_list',
            kwargs=self.test_object_view_kwargs
        )

    def _request_test_acl_permission_list_get_view(self):
        return self.get(
            viewname='acls:acl_permissions', kwargs={
                'acl_id': self.test_acl.pk
            }

        )

    def _request_test_acl_permission_add_view(self):
        return self.post(
            viewname='acls:acl_permissions', kwargs={
                'acl_id': self.test_acl.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_permission.stored_permission.pk
            }
        )

    def _request_test_acl_permission_remove_view(self):
        return self.post(
            viewname='acls:acl_permissions', kwargs={
                'acl_id': self.test_acl.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_permission.stored_permission.pk
            }
        )

    def _request_test_global_acl_list_view(self):
        return self.get(
            viewname='acls:global_acl_list',
        )



from mayan.apps.django_gpg.permissions import permission_key_sign
from mayan.apps.django_gpg.tests.mixins import KeyTestMixin
from mayan.apps.documents.models import DocumentFile
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase
from mayan.apps.documents.tests.literals import (
    TEST_DOCUMENT_PATH, TEST_SMALL_DOCUMENT_PATH
)

from ..models import DetachedSignature, EmbeddedSignature
from ..permissions import (
    permission_document_file_sign_detached,
    permission_document_file_sign_embedded,
    permission_document_file_signature_delete,
    permission_document_file_signature_download,
    permission_document_file_signature_upload,
    permission_document_file_signature_verify,
    permission_document_file_signature_view
)

from .literals import (
    TEST_SIGNED_DOCUMENT_COUNT, TEST_SIGNED_DOCUMENT_PATH,
    TEST_UNSIGNED_DOCUMENT_COUNT
)
from .mixins import (
    DetachedSignatureViewTestMixin, EmbeddedSignatureViewTestMixin,
    SignatureTestMixin, SignatureToolsViewTestMixin, SignatureViewTestMixin
)


class SignaturesViewTestCase(
    KeyTestMixin, SignatureTestMixin, SignatureViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_signature_delete_view_no_permission(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature_count = DetachedSignature.objects.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_view
        )

        response = self._request_test_document_file_signature_delete_view()
        self.assertEqual(response.status_code, 404)
        self.assertEqual(DetachedSignature.objects.count(), signature_count)

    def test_signature_delete_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature_count = DetachedSignature.objects.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_delete
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_view
        )

        response = self._request_test_document_file_signature_delete_view()
        self.assertEqual(response.status_code, 302)
        self.assertEqual(
            DetachedSignature.objects.count(), signature_count - 1
        )

    def test_trashed_document_signature_delete_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature_count = DetachedSignature.objects.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_delete
        )
        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_view
        )

        self.test_document.delete()

        response = self._request_test_document_file_signature_delete_view()
        self.assertEqual(response.status_code, 404)
        self.assertEqual(
            DetachedSignature.objects.count(), signature_count
        )

    def test_signature_detail_view_no_permission(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._upload_test_detached_signature()

        response = self._request_test_document_file_signature_details_view()
        self.assertEqual(response.status_code, 404)

    def test_signature_detail_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._create_test_key_public()
        self._upload_test_detached_signature()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_view
        )

        response = self._request_test_document_file_signature_details_view()
        self.assertContains(
            response=response, text=self.test_signature.signature_id,
            status_code=200
        )

    def test_trashed_document_signature_detail_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._create_test_key_public()
        self._upload_test_detached_signature()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_view
        )

        self.test_document.delete()

        response = self._request_test_document_file_signature_details_view()
        self.assertEqual(response.status_code, 404)

    def test_signature_list_view_no_permission(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        response = self._request_test_document_file_signature_list_view(
            document=self.test_document
        )
        self.assertEqual(response.status_code, 404)

    def test_signature_list_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_view
        )

        response = self._request_test_document_file_signature_list_view(
            document=self.test_document
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.context['object_list'].count(), 1)

    def test_trashed_document_signature_list_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_view
        )

        self.test_document.delete()

        response = self._request_test_document_file_signature_list_view(
            document=self.test_document
        )
        self.assertEqual(response.status_code, 404)


class DetachedSignaturesViewTestCase(
    KeyTestMixin, SignatureTestMixin, DetachedSignatureViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_detached_signature_create_view_with_no_permission(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )

    def test_detached_signature_create_view_with_document_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_sign_detached
        )

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )

    def test_detached_signature_create_view_with_key_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_key_private,
            permission=permission_key_sign
        )

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )

    def test_detached_signature_create_view_with_full_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_sign_detached
        )
        self.grant_access(
            obj=self.test_key_private,
            permission=permission_key_sign
        )

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count + 1
        )

    def test_trashed_document_detached_signature_create_view_with_full_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_sign_detached
        )
        self.grant_access(
            obj=self.test_key_private,
            permission=permission_key_sign
        )

        self.test_document.delete()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )

    def test_signature_download_view_no_permission(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        response = self._request_test_document_file_signature_download_view()
        self.assertEqual(response.status_code, 404)

    def test_signature_download_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_download
        )

        self.expected_content_types = ('application/octet-stream',)

        response = self._request_test_document_file_signature_download_view()

        with self.test_signature.signature_file as file_object:
            self.assert_download_response(
                response=response, content=file_object.read(),
            )

    def test_trashed_document_signature_download_view_with_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_download
        )

        self.test_document.delete()

        response = self._request_test_document_file_signature_download_view()
        self.assertEqual(response.status_code, 404)

    def test_signature_upload_view_no_permission(self):
        self.test_document_path = TEST_DOCUMENT_PATH

        signature_count = DetachedSignature.objects.count()

        self._upload_test_document()

        response = self._request_test_document_file_signature_upload_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(DetachedSignature.objects.count(), signature_count)

    def test_signature_upload_view_with_access(self):
        self.test_document_path = TEST_DOCUMENT_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_upload
        )

        signature_count = DetachedSignature.objects.count()

        response = self._request_test_document_file_signature_upload_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            DetachedSignature.objects.count(), signature_count + 1
        )

    def test_trashed_document_signature_upload_view_with_access(self):
        self.test_document_path = TEST_DOCUMENT_PATH
        self._upload_test_document()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_signature_upload
        )

        self.test_document.delete()

        signature_count = DetachedSignature.objects.count()

        response = self._request_test_document_file_signature_upload_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(DetachedSignature.objects.count(), signature_count)


class EmbeddedSignaturesViewTestCase(
    KeyTestMixin, EmbeddedSignatureViewTestMixin, GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_embedded_signature_create_view_with_no_permission(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )

    def test_embedded_signature_create_view_with_document_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_sign_embedded
        )

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )

    def test_embedded_signature_create_view_with_key_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_key_private,
            permission=permission_key_sign
        )

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )

    def test_embedded_signature_create_view_with_full_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_sign_embedded
        )
        self.grant_access(
            obj=self.test_key_private,
            permission=permission_key_sign
        )

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 302)
        self.assertTrue(
            str(self.test_document.file_latest.pk) in response.url
        )

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count + 1
        )

    def test_trashed_document_embedded_signature_create_view_with_full_access(self):
        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()
        self._create_test_key_private()

        signature_count = self.test_document.file_latest.signatures.count()

        self.grant_access(
            obj=self.test_document,
            permission=permission_document_file_sign_embedded
        )
        self.grant_access(
            obj=self.test_key_private,
            permission=permission_key_sign
        )

        self.test_document.delete()

        response = self._request_test_document_file_signature_create_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            self.test_document.file_latest.signatures.count(),
            signature_count
        )


class SignatureToolsViewTestCase(
    KeyTestMixin, SignatureTestMixin, SignatureToolsViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        class DetachedSignatureProxy(DetachedSignature):
            def save(self, *args, **kwargs):
                return super(DetachedSignature, self).save(*args, **kwargs)

            class Meta:
                proxy = True

        class EmbeddedSignatureProxy(EmbeddedSignature):
            def save(self, *args, **kwargs):
                return super(EmbeddedSignature, self).save(*args, **kwargs)

            class Meta:
                proxy = True

        cls.DetachedSignatureProxy = DetachedSignatureProxy
        cls.EmbeddedSignatureProxy = EmbeddedSignatureProxy

    def test_missing_signature_verify_view_no_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        old_hooks = DocumentFile._post_save_hooks
        DocumentFile._post_save_hooks = {}

        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        for count in range(TEST_UNSIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.test_document_path = TEST_SIGNED_DOCUMENT_PATH
        for count in range(TEST_SIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

        DocumentFile._post_save_hooks = old_hooks

        response = self._request_all_test_document_file_signature_verify_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

    def test_missing_signature_verify_view_with_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        old_hooks = DocumentFile._post_save_hooks
        DocumentFile._post_save_hooks = {}

        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        for count in range(TEST_UNSIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.test_document_path = TEST_SIGNED_DOCUMENT_PATH
        for count in range(TEST_SIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

        DocumentFile._post_save_hooks = old_hooks

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        response = self._request_all_test_document_file_signature_verify_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT
        )

    def test_trashed_document_missing_signature_verify_view_with_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        old_hooks = DocumentFile._post_save_hooks
        DocumentFile._post_save_hooks = {}

        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        for count in range(TEST_UNSIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.test_document_path = TEST_SIGNED_DOCUMENT_PATH
        for count in range(TEST_SIGNED_DOCUMENT_COUNT):
            self._upload_test_document()

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT + TEST_SIGNED_DOCUMENT_COUNT
        )

        DocumentFile._post_save_hooks = old_hooks

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        self.test_document.delete()

        response = self._request_all_test_document_file_signature_verify_view()

        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            EmbeddedSignature.objects.unsigned_document_files().count(),
            TEST_UNSIGNED_DOCUMENT_COUNT
        )

    def test_signature_refresh_view_no_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        self.test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature = self.DetachedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        signature = self.EmbeddedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        response = self._request_all_test_document_file_signature_refresh_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(
            DetachedSignature.objects.first().date_time,
            None
        )
        self.assertEqual(
            EmbeddedSignature.objects.first().date_time,
            None
        )

    def test_signature_refresh_view_with_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        self.test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature = self.DetachedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        signature = self.EmbeddedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        response = self._request_all_test_document_file_signature_refresh_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            DetachedSignature.objects.first().date_time,
            None
        )
        self.assertNotEqual(
            EmbeddedSignature.objects.first().date_time,
            None
        )

    def test_trashed_document_signature_refresh_view_with_permission(self):
        # Silence converter logging
        self._silence_logger(name='mayan.apps.converter.backends')

        self.test_document_path = TEST_SIGNED_DOCUMENT_PATH
        self._upload_test_document()

        self.test_document_path = TEST_SMALL_DOCUMENT_PATH
        self._upload_test_document()

        self._upload_test_detached_signature()

        signature = self.DetachedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        signature = self.EmbeddedSignatureProxy.objects.first()
        signature.date_time = None
        signature.save()

        self.test_documents[0].delete()
        self.test_documents[1].delete()

        self.grant_permission(
            permission=permission_document_file_signature_verify
        )

        response = self._request_all_test_document_file_signature_refresh_view()
        self.assertEqual(response.status_code, 302)

        self.assertNotEqual(
            DetachedSignature.objects.first().date_time,
            None
        )
        self.assertNotEqual(
            EmbeddedSignature.objects.first().date_time,
            None
        )



from django.apps import apps
from django.utils.encoding import force_text
from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.classes import DocumentCreateWizardStep
from mayan.apps.views.http import URL

from .forms import TagMultipleSelectionForm
from .models import Tag
from .permissions import permission_tag_attach


class DocumentCreateWizardStepTags(DocumentCreateWizardStep):
    form_class = TagMultipleSelectionForm
    label = _('Select tags')
    name = 'tag_selection'
    number = 2

    @classmethod
    def condition(cls, wizard):
        Tag = apps.get_model(app_label='tags', model_name='Tag')
        return Tag.objects.exists()

    @classmethod
    def get_form_kwargs(self, wizard):
        return {
            'help_text': _('Tags to be attached.'),
            'model': Tag,
            'permission': permission_tag_attach,
            'user': wizard.request.user
        }

    @classmethod
    def done(cls, wizard):
        result = {}
        cleaned_data = wizard.get_cleaned_data_for_step(cls.name)
        if cleaned_data:
            result['tags'] = [
                force_text(s=tag.pk) for tag in cleaned_data['tags']
            ]

        return result

    @classmethod
    def step_post_upload_process(cls, document, querystring=None):
        Tag = apps.get_model(app_label='tags', model_name='Tag')

        tag_id_list = URL(query_string=querystring).args.getlist('tags')

        for tag in Tag.objects.filter(pk__in=tag_id_list):
            tag.documents.add(document)


DocumentCreateWizardStep.register(step=DocumentCreateWizardStepTags)



from django.db import migrations


def operation_document_file_filename_copy(apps, schema_editor):
    cursor_main = schema_editor.connection.cursor()
    cursor_document_file = schema_editor.connection.cursor()

    query_document_file = '''
        UPDATE {documents_documentfile} SET {filename} = %s WHERE {documents_documentfile}.{id} = %s;
    '''.format(
        documents_documentfile=schema_editor.connection.ops.quote_name('documents_documentfile'),
        filename=schema_editor.connection.ops.quote_name('filename'),
        id=schema_editor.connection.ops.quote_name('id')
    )

    query = '''
        SELECT
            {documents_document}.{label},
            {documents_documentfile}.{id}
        FROM {documents_documentfile}
        INNER JOIN {documents_document} ON (
            {documents_documentfile}.{document_id} = {documents_document}.{id}
        )
    '''.format(
        document_id=schema_editor.connection.ops.quote_name('document_id'),
        documents_document=schema_editor.connection.ops.quote_name('documents_document'),
        documents_documentfile=schema_editor.connection.ops.quote_name('documents_documentfile'),
        id=schema_editor.connection.ops.quote_name('id'),
        label=schema_editor.connection.ops.quote_name('label')
    )

    cursor_main.execute(query)

    for row in cursor_main.fetchall():
        cursor_document_file.execute(
            query_document_file, row
        )


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0066_documentfile_filename'),
    ]

    operations = [
        migrations.RunPython(
            code=operation_document_file_filename_copy,
            reverse_code=migrations.RunPython.noop
        ),
    ]



import functools
import inspect
import logging

from django.core.exceptions import ImproperlyConfigured

from .backends.base import LockingBackend
from .exceptions import LockError

logger = logging.getLogger(name=__name__)


def locked_class_method(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if inspect.isgeneratorfunction(getattr(func, '__wrapped__', None)):
            raise ImproperlyConfigured(
                'This decorator does not support generators.'
            )

        _acquire_lock = kwargs.pop('_acquire_lock', True)
        lock_name = self._lock_manager_get_lock_name(*args, **kwargs)

        try:
            if _acquire_lock:
                logger.debug('trying to acquire lock: %s', lock_name)
                lock = LockingBackend.get_backend().acquire_lock(name=lock_name)
                logger.debug('acquired lock: %s', lock_name)
        except LockError:
            logger.debug('unable to obtain lock: %s' % lock_name)
            raise
        else:
            try:
                return func(self, *args, **kwargs)
            finally:
                if _acquire_lock:
                    lock.release()
    return wrapper


def acquire_lock_class_method(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if inspect.isgeneratorfunction(getattr(func, '__wrapped__', None)):
            raise ImproperlyConfigured(
                'This decorator does not support generators.'
            )

        _acquire_lock = kwargs.pop('_acquire_lock', True)
        lock_name = self._lock_manager_get_lock_name(*args, **kwargs)

        try:
            if _acquire_lock:
                logger.debug('trying to acquire lock: %s', lock_name)
                self._lock = LockingBackend.get_backend().acquire_lock(name=lock_name)
                logger.debug('acquired lock: %s', lock_name)
        except LockError:
            logger.debug('unable to obtain lock: %s' % lock_name)
            raise
        else:
            return func(self, *args, **kwargs)

    return wrapper


def release_lock_class_method(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if inspect.isgeneratorfunction(getattr(func, '__wrapped__', None)):
            raise ImproperlyConfigured(
                'This decorator does not support generators.'
            )

        _acquire_lock = kwargs.pop('_acquire_lock', True)

        try:
            return func(self, *args, **kwargs)
        finally:
            if _acquire_lock:
                self._lock.release()
    return wrapper



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    copyright_text='''
        Copyright (c) 2010-2015, Justin Quick
        All rights reserved.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted provided that the following conditions are
        met:

            * Redistributions of source code must retain the above copyright
              notice, this list of conditions and the following disclaimer.
            * Redistributions in binary form must reproduce the above
              copyright notice, this list of conditions and the following
              disclaimer in the documentation and/or other materials provided
              with the distribution.
            * Neither the name of the author nor the names of other
              contributors may be used to endorse or promote products derived
              from this software without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
        "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
        LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
        A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
        OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
        SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
        LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
        DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
        THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
        (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
        OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    ''', module=__name__, name='django-activity-stream',
    version_string='==0.8.0'
)



import logging

import yaml

from django.utils.translation import ugettext_lazy as _
from django.core.exceptions import ValidationError

from mayan.apps.common.serialization import yaml_load
from mayan.apps.documents.utils import parse_range
from mayan.apps.document_states.classes import WorkflowAction

from .models import ObjectLayer
from .transformations import BaseTransformation

__all__ = ('TransformationAddAction',)
logger = logging.getLogger(name=__name__)


class TransformationAddAction(WorkflowAction):
    fields = {
        'pages': {
            'label': _('Pages'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Pages to which the new transformations will be added. '
                    'Separate by commas and/or use a dashes for a ranges. '
                    'Leave blank to select all pages.'
                ), 'required': False
            }
        },
        'transformation_class': {
            'label': _('Transformation class'),
            'class': 'django.forms.ChoiceField', 'kwargs': {
                'choices': BaseTransformation.get_transformation_choices(
                    group_by_layer=True
                ), 'help_text': _(
                    'Type of transformation to add.'
                ), 'required': True
            }
        },
        'transformation_arguments': {
            'label': _('Transformation arguments'),
            'class': 'django.forms.CharField', 'kwargs': {
                'help_text': _(
                    'Enter the arguments for the transformation as a YAML '
                    'dictionary. ie: {"degrees": 180}'
                ), 'required': False
            }
        }
    }
    label = _('Add transformation')
    widgets = {
        'transformation_class': {
            'class': 'django.forms.widgets.Select', 'kwargs': {
                'attrs': {'class': 'select2'},
            }
        },
        'transformation_arguments': {
            'class': 'django.forms.widgets.Textarea', 'kwargs': {
                'attrs': {'rows': 5}
            }
        }
    }

    @classmethod
    def clean(cls, form_data, request):
        try:
            yaml_load(
                stream=form_data['action_data']['transformation_arguments']
            )
        except yaml.YAMLError:
            raise ValidationError(
                _(
                    '"%s" not a valid entry.'
                ) % form_data['action_data']['transformation_arguments']
            )

        return form_data

    def execute(self, context):
        if self.form_data['pages']:
            page_range = parse_range(astr=self.form_data['pages'])
            queryset = context['document'].pages.filter(
                page_number__in=page_range
            )
        else:
            queryset = context['document'].pages.all()

        transformation_class = BaseTransformation.get(name=self.form_data['transformation_class'])
        layer = transformation_class.get_assigned_layer()

        for document_page in queryset.all():
            object_layer, created = ObjectLayer.objects.get_for(
                obj=document_page, layer=layer
            )
            object_layer.transformations.create(
                name=transformation_class.name,
                arguments=self.form_data['transformation_arguments']
            )



from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0039_duplicateddocument'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='documentversion',
            options={
                'ordering': ('timestamp',),
                'verbose_name': 'Document version',
                'verbose_name_plural': 'Document version'
            },
        ),
    ]



from colorful.fields import RGBColorField

import whoosh
from whoosh import qparser  # NOQA Used to initialize the whoosh.fields module

from django.db import models

QUERY_OPERATION_AND = 1
QUERY_OPERATION_OR = 2
TERM_OPERATION_AND = 'AND'
TERM_OPERATION_OR = 'OR'
TERM_OPERATIONS = [TERM_OPERATION_AND, TERM_OPERATION_OR]
TERM_QUOTES = ['"', '\'']
TERM_NEGATION_CHARACTER = '-'
TERM_SPACE_CHARACTER = ' '

DJANGO_TO_WHOOSH_FIELD_MAP = {
    models.AutoField: {
        'field': whoosh.fields.ID(stored=True), 'transformation': str
    },
    models.CharField: {'field': whoosh.fields.TEXT},
    models.EmailField: {'field': whoosh.fields.TEXT},
    models.TextField: {'field': whoosh.fields.TEXT},
    models.UUIDField: {'field': whoosh.fields.TEXT, 'transformation': str},
    RGBColorField: {'field': whoosh.fields.TEXT},
}
WHOOSH_INDEX_DIRECTORY_NAME = 'whoosh'



from django.db import connection
from django.utils.translation import ugettext_lazy as _

from mayan.apps.dynamic_search.classes import SearchModel
from mayan.apps.views.literals import LIST_MODE_CHOICE_ITEM

from .permissions import (
    permission_document_file_view, permission_document_type_view,
    permission_document_version_view, permission_document_view
)


def transformation_format_uuid(term_string):
    if connection.vendor in ('mysql', 'sqlite'):
        return term_string.replace('-', '')
    else:
        return term_string


# Document

document_search = SearchModel(
    app_label='documents', default=True, label=_('Document'),
    list_mode=LIST_MODE_CHOICE_ITEM, model_name='DocumentSearchResult',
    permission=permission_document_view,
    serializer_path='mayan.apps.documents.serializers.document_serializers.DocumentSerializer'
)
document_search.add_proxy_model(app_label='documents', model_name='Document')

document_search.add_model_field(
    field='document_type__label', label=_('Document type label')
)
document_search.add_model_field(field='label')
document_search.add_model_field(field='description')
document_search.add_model_field(
    field='uuid', transformation_function=transformation_format_uuid
)
document_search.add_model_field(
    field='files__checksum', label=('Document file checksum')
)
document_search.add_model_field(
    field='files__filename', label=('Document file filename')
)
document_search.add_model_field(
    field='files__mimetype', label=('Document file MIME type')
)

# Document file

document_file_search = SearchModel(
    app_label='documents', label=_('Document file'),
    list_mode=LIST_MODE_CHOICE_ITEM, model_name='DocumentFileSearchResult',
    permission=permission_document_file_view,
    serializer_path='mayan.apps.documents.serializers.document_file_serializers.DocumentFileSerializer'
)
document_file_search.add_proxy_model(
    app_label='documents', model_name='DocumentFile'
)

document_file_search.add_model_field(
    field='document__document_type__label',
    label=_('Document type label')
)
document_file_search.add_model_field(
    field='document__label', label=_('Document label')
)
document_file_search.add_model_field(
    field='document__description', label=_('Document description')
)
document_file_search.add_model_field(field='checksum')
document_file_search.add_model_field(field='comment')
document_file_search.add_model_field(field='filename')
document_file_search.add_model_field(field='mimetype')

# Document file page

document_file_page_search = SearchModel(
    app_label='documents', label=_('Document file page'),
    list_mode=LIST_MODE_CHOICE_ITEM, model_name='DocumentFilePageSearchResult',
    permission=permission_document_file_view,
    serializer_path='mayan.apps.documents.serializers.document_file_serializers.DocumentFilePageSerializer'
)
document_file_page_search.add_proxy_model(
    app_label='documents', model_name='DocumentFilePage'
)

document_file_page_search.add_model_field(
    field='document_file__document__document_type__label',
    label=_('Document type label')
)
document_file_page_search.add_model_field(
    field='document_file__document__label', label=_('Document label')
)
document_file_page_search.add_model_field(
    field='document_file__document__description',
    label=_('Document description')
)
document_file_page_search.add_model_field(
    field='document_file__checksum', label=_('Document file checksum')
)
document_file_page_search.add_model_field(
    field='document_file__mimetype', label=_('Document file MIME type')
)

# Document type

document_type_search = SearchModel(
    app_label='documents', list_mode=LIST_MODE_CHOICE_ITEM,
    model_name='DocumentType', permission=permission_document_type_view,
    serializer_path='mayan.apps.documents.serializers.document_type_serializers.DocumentTypeSerializer'
)
document_type_search.add_model_field(field='label')

# Document version

document_version_search = SearchModel(
    app_label='documents', label=_('Document version'),
    list_mode=LIST_MODE_CHOICE_ITEM, model_name='DocumentVersionSearchResult',
    permission=permission_document_version_view,
    serializer_path='mayan.apps.documents.serializers.document_version_serializers.DocumentVersionSerializer'
)
document_version_search.add_proxy_model(
    app_label='documents', model_name='DocumentVersion'
)

document_version_search.add_model_field(field='comment')
document_version_search.add_model_field(
    field='document__document_type__label',
    label=_('Document type label')
)
document_version_search.add_model_field(
    field='document__label', label=_('Document label')
)
document_version_search.add_model_field(
    field='document__description', label=_('Document description')
)

# Document version page

document_version_page_search = SearchModel(
    app_label='documents', label=_('Document version page'),
    list_mode=LIST_MODE_CHOICE_ITEM,
    model_name='DocumentVersionPageSearchResult',
    permission=permission_document_version_view,
    serializer_path='mayan.apps.documents.serializers.document_version_serializers.DocumentVersionPageSerializer'
)
document_version_page_search.add_proxy_model(
    app_label='documents', model_name='DocumentVersionPage'
)

document_version_page_search.add_model_field(
    field='document_version__document__document_type__label',
    label=_('Document type label')
)
document_version_page_search.add_model_field(
    field='document_version__comment', label=_('Document version comment')
)
document_version_page_search.add_model_field(
    field='document_version__document__label', label=_('Document label')
)
document_version_page_search.add_model_field(
    field='document_version__document__description',
    label=_('Document description')
)



from actstream.models import Action

from mayan.apps.documents.tests.base import GenericDocumentTestCase

from ..events import (
    event_file_metadata_document_file_finish,
    event_file_metadata_document_file_submit
)


class FileMetadataEventsTestCase(GenericDocumentTestCase):
    def test_document_file_finish_event(self):
        Action.objects.all().delete()
        self.test_document.file_latest.submit_for_file_metadata_processing()

        # Get the most recent action
        action = Action.objects.order_by('-timestamp').first()

        self.assertEqual(
            action.target, self.test_document.file_latest
        )
        self.assertEqual(
            action.verb, event_file_metadata_document_file_finish.id
        )

    def test_document_file_submit_event(self):
        Action.objects.all().delete()
        self.test_document.file_latest.submit_for_file_metadata_processing()

        # Get the oldest action
        action = Action.objects.order_by('-timestamp').last()

        self.assertEqual(
            action.target, self.test_document.file_latest
        )
        self.assertEqual(
            action.verb, event_file_metadata_document_file_submit.id
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.permissions import PermissionNamespace

namespace = PermissionNamespace(label=_('Events'), name='events')

permission_events_export = namespace.add_permission(
    label=_('Export the events of an object'), name='events_export'
)
permission_events_view = namespace.add_permission(
    label=_('Access the events of an object'), name='events_view'
)



from django.apps import apps
from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import get_content_type_kwargs_factory

from .icons import (
    icon_current_user_events, icon_current_user_events_export,
    icon_event_types_subscriptions_list, icon_events_for_object,
    icon_events_for_object_export, icon_events_list,
    icon_events_list_export, icon_notification_mark_read,
    icon_notification_mark_read_all,
    icon_object_event_types_user_subcriptions_list,
    icon_user_notifications_list
)
from .permissions import permission_events_view


def get_unread_notification_count(context):
    Notification = apps.get_model(
        app_label='events', model_name='Notification'
    )
    if context.request.user.is_authenticated:
        return Notification.objects.filter(
            user=context.request.user
        ).filter(read=False).count()


link_current_user_events = Link(
    icon=icon_current_user_events, text=_('My events'),
    view='events:current_user_events'
)
link_current_user_events_export = Link(
    icon=icon_current_user_events_export, text=_('Export events'),
    view='events:current_user_events_export'
)
link_events_details = Link(
    text=_('Events'), view='events:events_list'
)
link_events_for_object = Link(
    icon=icon_events_for_object,
    kwargs=get_content_type_kwargs_factory(variable_name='resolved_object'),
    permissions=(permission_events_view,), text=_('Events'),
    view='events:events_for_object'
)
link_events_for_object_export = Link(
    icon=icon_events_for_object_export,
    kwargs=get_content_type_kwargs_factory(variable_name='resolved_object'),
    permissions=(permission_events_view,), text=_('Export events'),
    view='events:events_for_object_export'
)
link_events_list = Link(
    icon=icon_events_list, text=_('Events'), view='events:events_list'
)
link_events_list_export = Link(
    icon=icon_events_list_export, text=_('Export events'),
    view='events:events_list_export'
)
link_event_types_subscriptions_list = Link(
    icon=icon_event_types_subscriptions_list,
    text=_('Event subscriptions'),
    view='events:event_types_user_subcriptions_list'
)
link_notification_mark_read = Link(
    args='object.pk', icon=icon_notification_mark_read,
    text=_('Mark as seen'), view='events:notification_mark_read'
)
link_notification_mark_read_all = Link(
    icon=icon_notification_mark_read_all, text=_('Mark all as seen'),
    view='events:notification_mark_read_all'
)
link_object_event_types_user_subcriptions_list = Link(
    icon=icon_object_event_types_user_subcriptions_list,
    kwargs=get_content_type_kwargs_factory(variable_name='resolved_object'),
    permissions=(permission_events_view,), text=_('Subscriptions'),
    view='events:object_event_types_user_subcriptions_list',
)
link_user_notifications_list = Link(
    badge_text=get_unread_notification_count,
    icon=icon_user_notifications_list, text='',
    view='events:user_notifications_list'
)



import logging

from django.utils.termcolors import colorize

from .literals import FORMATTER_PALETTE


class ColorFormatter(logging.Formatter):
    def format(self, record):
        record.msg = colorize(
            text=record.msg, **FORMATTER_PALETTE.get(record.levelname, {})
        )
        return super().format(record)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.apps import MayanAppConfig
from mayan.apps.common.menus import menu_secondary
from mayan.apps.navigation.classes import SourceColumn

from .links import link_object_error_list_clear
from .mixins import LoggingAppConfigMixin


class LoggingApp(LoggingAppConfigMixin, MayanAppConfig):
    app_namespace = 'logging'
    app_url = 'logging'
    has_tests = True
    name = 'mayan.apps.logging'
    verbose_name = _('Logging')

    def ready(self, *args, **kwargs):
        super().ready(*args, **kwargs)

        ErrorLogPartitionEntry = self.get_model(
            model_name='ErrorLogPartitionEntry'
        )

        SourceColumn(
            attribute='datetime', is_identifier=True, is_sortable=True,
            source=ErrorLogPartitionEntry
        )
        SourceColumn(
            attribute='text', include_label=True, is_sortable=True,
            source=ErrorLogPartitionEntry
        )

        menu_secondary.bind_links(
            links=(link_object_error_list_clear,), sources=(
                'logging:object_error_list',
            )
        )



from django.core import management
from django.core.management.base import CommandError

from ...signals import signal_perform_upgrade, signal_post_upgrade, signal_pre_upgrade


class Command(management.BaseCommand):
    help = 'Performs the required steps after a version upgrade.'

    def add_arguments(self, parser):
        parser.add_argument(
            '--no-dependencies', action='store_true', dest='no_dependencies',
            help='Don\'t download dependencies.',
        )

    def handle(self, *args, **options):
        try:
            signal_pre_upgrade.send(sender=self)
        except Exception as exception:
            raise CommandError(
                'Error during signal_pre_upgrade signal: %s, %s' % (
                    exception, type(exception)
                )
            )

        if not options.get('no_dependencies', False):
            management.call_command(
                command_name='installdependencies'
            )
            management.call_command(
                command_name='preparestatic', interactive=False
            )

        try:
            signal_perform_upgrade.send(sender=self)
        except Exception as exception:
            raise CommandError(
                'Error during signal_perform_upgrade signal; %s, %s' % (
                    exception, type(exception)
                )
            )

        try:
            signal_post_upgrade.send(sender=self)
        except Exception as exception:
            raise CommandError(
                'Error during signal_post_upgrade signal; %s, %s' % (
                    exception, type(exception)
                )
            )



from mayan.apps.documents.tests.base import GenericDocumentTestCase

from .mixins import SmartLinkTestMixin


class SmartLinkTestCase(SmartLinkTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()

        self._create_test_document_stub()
        self._create_test_smart_link(add_test_document_type=True)

    def test_smart_link_dynamic_label(self):
        self.assertEqual(
            self.test_smart_link.get_dynamic_label(document=self.test_document),
            self.test_document.label
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.events.classes import EventTypeNamespace

namespace = EventTypeNamespace(label=_('Mailing'), name='mailing')

event_email_sent = namespace.add_event_type(
    label=_('Email sent'), name='email_send'
)



from django.utils.translation import ugettext_lazy as _

from .events import event_file_metadata_document_file_submit
from .tasks import task_process_document_file


def method_document_submit(self):
    latest_file = self.file_latest
    # Don't error out if document has no file
    if latest_file:
        latest_file.submit_for_file_metadata_processing()


def method_document_file_submit(self):
    event_file_metadata_document_file_submit.commit(
        action_object=self.document, target=self
    )

    task_process_document_file.apply_async(
        kwargs={
            'document_file_id': self.pk,
        }
    )


def method_get_document_file_metadata(self, dotted_name):
    latest_file = self.file_latest
    # Don't error out if document has no file
    if latest_file:
        return latest_file.get_file_metadata(
            dotted_name=dotted_name
        )


method_get_document_file_metadata.short_description = _(
    'get_file_metadata(< file metadata dotted path >)'
)
method_get_document_file_metadata.help_text = _(
    'Return the specified document file metadata entry.'
)


def method_get_document_file_file_metadata(self, dotted_name):
    parts = dotted_name.split('__', 1)

    if len(parts) < 2:
        return

    driver_internal_name = parts[0]
    key = parts[1].replace('.', '_')

    try:
        document_driver = self.file_metadata_drivers.get(
            driver__internal_name=driver_internal_name
        )
    except self.file_metadata_drivers.model.DoesNotExist:
        return
    else:
        try:
            return document_driver.entries.get(key=key).value
        except document_driver.entries.model.DoesNotExist:
            return


method_get_document_file_file_metadata.help_text = _(
    'Return the specified document file file metadata entry.'
)



from django.utils.html import format_html


def jstree_data(node, selected_node):
    result = []

    result.append('{')
    result.append(format_html('"text": "{}",', node.label))
    result.append(
        '"state": {{ "opened": true, "selected": {} }},'.format(
            'true' if node == selected_node else 'false'
        )
    )
    result.append(
        '"data": {{ "href": "{}" }},'.format(node.get_absolute_url())
    )

    children = node.get_children().order_by('label',)

    if children:
        result.append('"children" : [')
        for child in children:
            result.extend(jstree_data(node=child, selected_node=selected_node))

        result.append(']')

    result.append('},')

    return result



from django.utils.translation import ugettext_lazy as _

from mayan.apps.navigation.classes import Link
from mayan.apps.navigation.utils import get_cascade_condition

from .classes import LayerLink
from .icons import (
    icon_asset_create, icon_asset_delete, icon_asset_edit, icon_asset_list,
    icon_transformation_delete, icon_transformation_edit,
    icon_transformation_select
)
from .layers import layer_saved_transformations
from .permissions import (
    permission_asset_create, permission_asset_delete,
    permission_asset_edit, permission_asset_view
)
from .transformations import BaseTransformation


def conditional_active(context, resolved_link):
    return resolved_link.link.view == resolved_link.current_view_name and context.get('layer_name', None) == resolved_link.link.layer_name


def condition_valid_transformation_and_arguments(context):
    try:
        transformation = BaseTransformation.get(name=context['object'].name)
    except KeyError:
        return False
    else:
        return transformation.arguments


link_asset_create = Link(
    icon=icon_asset_create, permissions=(permission_asset_create,),
    text=_('Create asset'), view='converter:asset_create'
)
link_asset_multiple_delete = Link(
    icon=icon_asset_delete, tags='dangerous', text=_('Delete'),
    view='converter:asset_multiple_delete'
)
link_asset_single_delete = Link(
    args='object.pk', icon=icon_asset_delete,
    permissions=(permission_asset_delete,), tags='dangerous',
    text=_('Delete'), view='converter:asset_single_delete'
)
link_asset_edit = Link(
    args='object.pk', icon=icon_asset_edit,
    permissions=(permission_asset_edit,), text=_('Edit'),
    view='converter:asset_edit'
)
link_asset_list = Link(
    condition=get_cascade_condition(
        app_label='converter', model_name='Asset',
        object_permission=permission_asset_view,
        view_permission=permission_asset_create,
    ), icon=icon_asset_list, text=_('Assets'),
    view='converter:asset_list'
)

link_transformation_delete = LayerLink(
    action='delete', kwargs={
        'layer_name': 'layer_name', 'transformation_id': 'resolved_object.pk'
    }, icon=icon_transformation_delete,
    layer=layer_saved_transformations, tags='dangerous', text=_('Delete'),
    view='converter:transformation_delete'
)
link_transformation_edit = LayerLink(
    action='edit', condition=condition_valid_transformation_and_arguments,
    kwargs={
        'layer_name': 'layer_name', 'transformation_id': 'resolved_object.pk'
    }, icon=icon_transformation_edit,
    layer=layer_saved_transformations, text=_('Edit'),
    view='converter:transformation_edit'
)
link_transformation_list = LayerLink(
    action='list', conditional_active=conditional_active,
    layer=layer_saved_transformations, text=_('Transformations'),
    view='converter:transformation_list'
)
link_transformation_select = LayerLink(
    action='select', icon=icon_transformation_select,
    layer=layer_saved_transformations, text=_('Select new transformation'),
    view='converter:transformation_select'
)



from django.conf.urls import url

from .api_views import (
    APIDocumentFileDetachedSignatureDetailView,
    APIDocumentFileDetachedSignatureListView,
    APIDocumentFileDetachedSignatureUploadView,
    APIDocumentFileEmbeddedSignatureDetailView,
    APIDocumentFileEmbeddedSignatureListView,
    APIDocumentFileSignDetachedView, APIDocumentFileSignEmbeddedView
)

from .views import (
    AllDocumentSignatureRefreshView, AllDocumentSignatureVerifyView,
    DocumentFileDetachedSignatureCreateView,
    DocumentFileEmbeddedSignatureCreateView, DocumentFileSignatureDeleteView,
    DocumentFileSignatureDetailView, DocumentFileSignatureDownloadView,
    DocumentFileSignatureListView, DocumentFileSignatureUploadView
)

urlpatterns = [
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/signatures/$',
        name='document_file_signature_list',
        view=DocumentFileSignatureListView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/signatures/detached/create/$',
        name='document_file_signature_detached_create',
        view=DocumentFileDetachedSignatureCreateView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/signatures/detached/upload/$',
        name='document_file_signature_upload',
        view=DocumentFileSignatureUploadView.as_view()
    ),
    url(
        regex=r'^documents/files/(?P<document_file_id>\d+)/signatures/embedded/create/$',
        name='document_file_signature_embedded_create',
        view=DocumentFileEmbeddedSignatureCreateView.as_view()
    ),
    url(
        regex=r'^signatures/(?P<signature_id>\d+)/delete/$',
        name='document_file_signature_delete',
        view=DocumentFileSignatureDeleteView.as_view()
    ),
    url(
        regex=r'^signatures/(?P<signature_id>\d+)/details/$',
        name='document_file_signature_details',
        view=DocumentFileSignatureDetailView.as_view()
    ),
    url(
        regex=r'^signatures/(?P<signature_id>\d+)/download/$',
        name='document_file_signature_download',
        view=DocumentFileSignatureDownloadView.as_view()
    ),
    url(
        regex=r'^tools/all/document/file/signature/refresh/$',
        name='all_document_file_signature_refresh',
        view=AllDocumentSignatureRefreshView.as_view()
    ),
    url(
        regex=r'^tools/all/document/file/signature/verify/$',
        name='all_document_file_signature_verify',
        view=AllDocumentSignatureVerifyView.as_view()
    )
]

api_urls = [
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/signatures/detached/$',
        name='document-file-signature-detached-list',
        view=APIDocumentFileDetachedSignatureListView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/signatures/detached/sign/$',
        name='document-file-signature-detached-sign',
        view=APIDocumentFileSignDetachedView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/signatures/detached/upload/$',
        name='document-file-signature-detached-upload',
        view=APIDocumentFileDetachedSignatureUploadView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/signatures/detached/(?P<detached_signature_id>[0-9]+)/$',
        name='detachedsignature-detail',
        view=APIDocumentFileDetachedSignatureDetailView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/signatures/embedded/$',
        name='document-file-signature-embedded-list',
        view=APIDocumentFileEmbeddedSignatureListView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/signatures/embedded/sign/$',
        name='document-file-signature-embedded-sign',
        view=APIDocumentFileSignEmbeddedView.as_view()
    ),
    url(
        regex=r'^documents/(?P<document_id>[0-9]+)/files/(?P<document_file_id>[0-9]+)/signatures/embedded/(?P<embedded_signature_id>[0-9]+)/$',
        name='embeddedsignature-detail',
        view=APIDocumentFileEmbeddedSignatureDetailView.as_view()
    ),
]



from mayan.settings.literals import (
    MAYAN_WORKER_A_CONCURRENCY, MAYAN_WORKER_A_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_A_MAX_TASKS_PER_CHILD, MAYAN_WORKER_B_CONCURRENCY,
    MAYAN_WORKER_B_MAX_MEMORY_PER_CHILD, MAYAN_WORKER_B_MAX_TASKS_PER_CHILD,
    MAYAN_WORKER_C_CONCURRENCY, MAYAN_WORKER_C_MAX_MEMORY_PER_CHILD,
    MAYAN_WORKER_C_MAX_TASKS_PER_CHILD, MAYAN_WORKER_D_CONCURRENCY,
    MAYAN_WORKER_D_MAX_MEMORY_PER_CHILD, MAYAN_WORKER_D_MAX_TASKS_PER_CHILD
)

from .classes import Worker

worker_a = Worker(
    concurrency=MAYAN_WORKER_A_CONCURRENCY,
    maximum_memory_per_child=MAYAN_WORKER_A_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_A_MAX_TASKS_PER_CHILD,
    name='worker_a', nice_level=0
)
worker_b = Worker(
    concurrency=MAYAN_WORKER_B_CONCURRENCY,
    maximum_memory_per_child=MAYAN_WORKER_B_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_B_MAX_TASKS_PER_CHILD,
    name='worker_b', nice_level=2
)
worker_c = Worker(
    concurrency=MAYAN_WORKER_C_CONCURRENCY,
    maximum_memory_per_child=MAYAN_WORKER_C_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_C_MAX_TASKS_PER_CHILD,
    name='worker_c', nice_level=15
)
worker_d = Worker(
    concurrency=MAYAN_WORKER_D_CONCURRENCY,
    maximum_memory_per_child=MAYAN_WORKER_D_MAX_MEMORY_PER_CHILD,
    maximum_tasks_per_child=MAYAN_WORKER_D_MAX_TASKS_PER_CHILD,
    name='worker_d', nice_level=18
)



from django.db.models import Q

from ..classes import Permission, PermissionNamespace
from ..models import Role

from .literals import (
    TEST_CASE_ROLE_LABEL, TEST_PERMISSION_LABEL, TEST_PERMISSION_LABEL_2,
    TEST_PERMISSION_NAME, TEST_PERMISSION_NAME_2, TEST_PERMISSION_NAMESPACE_LABEL,
    TEST_PERMISSION_NAMESPACE_LABEL_2, TEST_PERMISSION_NAMESPACE_NAME,
    TEST_PERMISSION_NAMESPACE_NAME_2, TEST_ROLE_LABEL, TEST_ROLE_LABEL_EDITED
)


class GroupRoleAddRemoveViewTestMixin:
    def _request_test_group_role_add_remove_get_view(self):
        return self.get(
            viewname='permissions:group_roles', kwargs={
                'group_id': self.test_group.pk
            }
        )

    def _request_test_group_role_add_view(self):
        return self.post(
            viewname='permissions:group_roles', kwargs={
                'group_id': self.test_group.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_role.pk
            }
        )

    def _request_test_group_role_remove_view(self):
        return self.post(
            viewname='permissions:group_roles', kwargs={
                'group_id': self.test_group.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_role.pk
            }
        )


class PermissionAPIViewTestMixin:
    def _request_permissions_list_api_view(self):
        return self.get(viewname='rest_api:permission-list')


class PermissionTestMixin:
    def _create_test_permission(self):
        self.test_permission_namespace = PermissionNamespace(
            label=TEST_PERMISSION_NAMESPACE_LABEL,
            name=TEST_PERMISSION_NAMESPACE_NAME
        )
        self.test_permission = self.test_permission_namespace.add_permission(
            label=TEST_PERMISSION_LABEL,
            name=TEST_PERMISSION_NAME
        )

    def _create_test_permission_2(self):
        self.test_permission_namespace_2 = PermissionNamespace(
            label=TEST_PERMISSION_NAMESPACE_LABEL_2,
            name=TEST_PERMISSION_NAMESPACE_NAME_2
        )
        self.test_permission_2 = self.test_permission_namespace_2.add_permission(
            label=TEST_PERMISSION_LABEL_2,
            name=TEST_PERMISSION_NAME_2
        )


class PermissionTestCaseMixin:
    def setUp(self):
        super().setUp()
        Permission.invalidate_cache()


class RoleAPIViewTestMixin:
    def _request_test_role_create_api_view(self, extra_data=None):
        pk_list = list(Role.objects.values_list('pk', flat=True))

        data = {
            'label': TEST_ROLE_LABEL
        }

        if extra_data:
            data.update(extra_data)

        response = self.post(
            viewname='rest_api:role-list', data=data
        )

        try:
            self.test_role = Role.objects.get(
                ~Q(pk__in=pk_list)
            )
        except Role.DoesNotExist:
            self.test_role = None

        return response

    def _request_test_role_delete_api_view(self):
        return self.delete(
            viewname='rest_api:role-detail', kwargs={
                'role_id': self.test_role.pk
            }
        )

    def _request_test_role_edit_api_view(
        self, extra_data=None, request_type='patch'
    ):
        data = {
            'label': TEST_ROLE_LABEL_EDITED
        }

        if extra_data:
            data.update(extra_data)

        return getattr(self, request_type)(
            viewname='rest_api:role-detail', kwargs={
                'role_id': self.test_role.pk
            }, data=data
        )

    def _request_test_role_edit_api_patch_view_extra_data(self):
        extra_data = {
            'groups_pk_list': '{}'.format(self.test_group.pk),
            'permissions_pk_list': '{}'.format(self.test_permission.pk)
        }
        return self._request_test_role_edit_api_view(
            extra_data=extra_data, request_type='patch'
        )

    def _request_test_role_edit_api_put_view_extra_data(self):
        extra_data = {
            'groups_pk_list': '{}'.format(self.test_group.pk),
            'permissions_pk_list': '{}'.format(self.test_permission.pk)
        }
        return self._request_test_role_edit_api_view(
            extra_data=extra_data, request_type='put'
        )

    def _request_test_role_list_api_view(self):
        return self.get(viewname='rest_api:role-list')


class RoleGroupAPIViewTestMixin:
    def _request_test_role_group_add_api_view(self):
        return self.post(
            viewname='rest_api:role-group-add', kwargs={
                'role_id': self.test_role.pk
            }, data={'group': self.test_group.pk}
        )

    def _request_test_role_group_list_api_view(self):
        return self.get(
            viewname='rest_api:role-group-list', kwargs={
                'role_id': self.test_role.pk
            }
        )

    def _request_test_role_group_remove_api_view(self):
        return self.post(
            viewname='rest_api:role-group-remove', kwargs={
                'role_id': self.test_role.pk
            }, data={'group': self.test_group.pk}
        )


class RolePermissionAPIViewTestMixin:
    def _request_test_role_permission_add_api_view(self):
        return self.post(
            viewname='rest_api:role-permission-add', kwargs={
                'role_id': self.test_role.pk
            }, data={'permission': self.test_permission.pk}
        )

    def _request_test_role_permission_list_api_view(self):
        return self.get(
            viewname='rest_api:role-permission-list', kwargs={
                'role_id': self.test_role.pk
            }
        )

    def _request_test_role_permission_remove_api_view(self):
        return self.post(
            viewname='rest_api:role-permission-remove', kwargs={
                'role_id': self.test_role.pk
            }, data={'permission': self.test_permission.pk}
        )


class RoleTestCaseMixin:
    def setUp(self):
        super().setUp()
        if hasattr(self, '_test_case_group'):
            self.create_role()

    def create_role(self):
        self._test_case_role = Role.objects.create(label=TEST_CASE_ROLE_LABEL)

    def grant_permission(self, permission):
        self._test_case_role.grant(permission=permission)

    def revoke_permission(self, permission):
        self._test_case_role.revoke(permission=permission)


class RoleGroupAddRemoveViewTestMixin:
    def _request_test_role_group_add_remove_get_view(self):
        return self.get(
            viewname='permissions:role_groups', kwargs={
                'role_id': self.test_role.pk
            }
        )

    def _request_test_role_group_add_view(self):
        return self.post(
            viewname='permissions:role_groups', kwargs={
                'role_id': self.test_role.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_group.pk
            }
        )

    def _request_test_role_group_remove_view(self):
        return self.post(
            viewname='permissions:role_groups', kwargs={
                'role_id': self.test_role.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_group.pk
            }
        )


class RolePermissionAddRemoveViewTestMixin:
    def _request_test_role_permission_add_remove_get_view(self):
        return self.get(
            viewname='permissions:role_permissions', kwargs={
                'role_id': self.test_role.pk
            }
        )

    def _request_test_role_permission_add_view(self):
        return self.post(
            viewname='permissions:role_permissions', kwargs={
                'role_id': self.test_role.pk,
            }, data={
                'available-submit': 'true',
                'available-selection': self.test_permission.stored_permission.pk
            }
        )

    def _request_test_role_permission_remove_view(self):
        return self.post(
            viewname='permissions:role_permissions', kwargs={
                'role_id': self.test_role.pk,
            }, data={
                'added-submit': 'true',
                'added-selection': self.test_permission.stored_permission.pk
            }
        )


class RoleTestMixin:
    def setUp(self):
        super().setUp()
        self.test_roles = []

    def _create_test_role(self, add_groups=None):
        total_test_roles = len(self.test_roles)
        label = '{}_{}'.format(TEST_ROLE_LABEL, total_test_roles)

        self.test_role = Role.objects.create(label=label)

        self.test_roles.append(self.test_role)

        for group in add_groups or []:
            self.test_role.groups.add(group)


class RoleViewTestMixin:
    def _request_test_role_create_view(self):
        # Typecast to list to force queryset evaluation
        values = list(Role.objects.values_list('pk', flat=True))

        response = self.post(
            viewname='permissions:role_create', data={
                'label': TEST_ROLE_LABEL,
            }
        )

        self.test_role = Role.objects.exclude(pk__in=values).first()

        return response

    def _request_test_role_delete_view(self):
        return self.post(
            viewname='permissions:role_delete', kwargs={
                'role_id': self.test_role.pk
            }
        )

    def _request_test_role_edit_view(self):
        return self.post(
            viewname='permissions:role_edit', kwargs={
                'role_id': self.test_role.pk
            }, data={
                'label': TEST_ROLE_LABEL_EDITED,
            }
        )

    def _request_test_role_list_view(self):
        return self.get(viewname='permissions:role_list')
